"""Common test fixtures for grammar testing."""

from typing import Any, Dict, List, Tuple, Type

import pytest

from sqlfluff.core.parser import KeywordSegment, StringParser
from sqlfluff.core.parser.context import ParseContext
from sqlfluff.core.parser.grammar.base import BaseGrammar
from sqlfluff.core.parser.types import ParseMode


@pytest.fixture(scope="function")
def structural_parse_mode_test(generate_test_segments, fresh_ansi_dialect):
    """Test the structural function of a grammar in various parse modes.

    This helper fixture is designed to modularise grammar tests.
    """

    def _structural_parse_mode_test(
        test_segment_seeds: List[str],
        grammar_class: Type[BaseGrammar],
        grammar_argument_seeds: List[str],
        grammar_terminator_seeds: List[str],
        grammar_kwargs: Dict[str, Any],
        parse_mode: ParseMode,
        input_slice: slice,
        output_tuple: Tuple[Any, ...],
    ):
        segments = generate_test_segments(test_segment_seeds)
        # Dialect is required here only to have access to bracket segments.
        ctx = ParseContext(dialect=fresh_ansi_dialect)

        # NOTE: We pass terminators using kwargs rather than directly because some
        # classes don't support it (e.g. Bracketed).
        if grammar_terminator_seeds:
            grammar_kwargs["terminators"] = [
                StringParser(e, KeywordSegment) for e in grammar_terminator_seeds
            ]

        _seq = grammar_class(
            *(StringParser(e, KeywordSegment) for e in grammar_argument_seeds),
            parse_mode=parse_mode,
            **grammar_kwargs,
        )
        _start = input_slice.start or 0
        _stop = input_slice.stop or len(segments)
        _match = _seq.match(segments[:_stop], _start, ctx)
        # If we're expecting an output tuple, assert the match is truthy.
        if output_tuple:
            assert _match
        _result = tuple(
            e.to_tuple(show_raw=True, code_only=False, include_meta=True)
            for e in _match.apply(segments)
        )
        assert _result == output_tuple

    # Return the function
    return _structural_parse_mode_test



"""Tests the combination of LT02 and LT11.

LT02: Indentation not consistent with previous lines
LT11: Set operators should be surrounded by newlines

Auto fix of LT11 does not insert correct indentation but just Newlines. It relies on
LT02 to sort out the indentation later. This is what is getting tested here.
"""

import sqlfluff


def test__rules__std_LT02_LT11_union_all_in_subquery_lint():
    """Verify a that LT11 reports lint errors in subqueries."""
    sql = (
        "SELECT * FROM (\n"
        "    SELECT 'g' UNION ALL\n"
        "    SELECT 'h'\n"
        "    UNION ALL SELECT 'j'\n"
        ")\n"
    )
    result = sqlfluff.lint(sql)

    assert "LT11" in [r["code"] for r in result]


def test__rules__std_LT02_LT11_union_all_in_subquery_fix():
    """Verify combination of rules LT02 and LT11 produces a correct indentation."""
    sql = (
        "SELECT c FROM (\n"
        "    SELECT 'g' UNION ALL\n"
        "    SELECT 'h'\n"
        "    UNION ALL SELECT 'j'\n"
        ")\n"
    )
    fixed_sql = (
        "SELECT c FROM (\n"
        "    SELECT 'g'\n"
        "    UNION ALL\n"
        "    SELECT 'h'\n"
        "    UNION ALL\n"
        "    SELECT 'j'\n"
        ")\n"
    )
    result = sqlfluff.fix(sql)
    assert result == fixed_sql



"""Showing how to use the python API to filter SQL files.

This example shows how to use the Linter class to parse,
and then process, SQL scripts. The methods shown can be
very powerful for searching and filtering SQL scripts.
"""

from sqlfluff.core import Linter

# Let's make a SQL script with a few statements in it
sql = """
SELECT 1;
CREATE TABLE tbl (a int, b varchar);
SELECT 100;
INSERT INTO tbl VALUES (1, 'abc');
SELECT 10000;
"""
print("SQL Script: ", repr(sql))

# Call .parse_string() directly on the Linter object.
# The result is a ParsedString object.
linter = Linter(dialect="ansi")
parsed = linter.parse_string(sql)

# Get access to the parsed syntax tree.
tree = parsed.tree

# The root element of the parse tree should be a file segment
assert tree.is_type("file")

# The children of that segment are a mixture of statements
# and separators. Each of those are also segments with the same
# available methods. We can make a list of their raw representations.
sections = [(segment.get_type(), segment.raw) for segment in tree.segments]
assert sections == [
    ("newline", "\n"),
    ("statement", "SELECT 1"),
    ("statement_terminator", ";"),
    ("newline", "\n"),
    ("statement", "CREATE TABLE tbl (a int, b varchar)"),
    ("statement_terminator", ";"),
    ("newline", "\n"),
    ("statement", "SELECT 100"),
    ("statement_terminator", ";"),
    ("newline", "\n"),
    ("statement", "INSERT INTO tbl VALUES (1, 'abc')"),
    ("statement_terminator", ";"),
    ("newline", "\n"),
    ("statement", "SELECT 10000"),
    ("statement_terminator", ";"),
    ("newline", "\n"),
    ("end_of_file", ""),  # There's a final "end of file segment"
]

# There are a few search methods available for filtering, in particular when
# looking for segments of a specific type, when they might not be direct
# siblings of the parent, you can use .recursive_crawl().
# NOTE: In a performance sensitive application, we recommend you limit the
# search depth by setting the `no_recursive_seg_type` option.
select_statements = tree.recursive_crawl("select_statement")
selects = [(segment.get_type(), segment.raw) for segment in select_statements]
assert selects == [
    ("select_statement", "SELECT 1"),
    ("select_statement", "SELECT 100"),
    ("select_statement", "SELECT 10000"),
]

# We could make a file, which only includes these statements by adding back in
# the statement terminators:
filtered_script = ";\n".join(select[1] for select in selects)
assert filtered_script == "SELECT 1;\nSELECT 100;\nSELECT 10000"
print("Filtered SQL Script: ", repr(filtered_script))



"""Methods for loading config from pyproject.toml files."""

import sys
from typing import Any, Dict, TypeVar

if sys.version_info >= (3, 11):
    import tomllib
else:  # pragma: no cover
    import toml as tomllib

from sqlfluff.core.helpers.dict import (
    NestedDictRecord,
    iter_records_from_nested_dict,
    records_to_nested_dict,
)
from sqlfluff.core.types import ConfigMappingType

T = TypeVar("T")


def _condense_rule_record(record: NestedDictRecord[T]) -> NestedDictRecord[T]:
    """Helper function to condense the rule section of a toml config."""
    key, value = record
    if len(key) > 2:
        key = (".".join(key[:-1]), key[-1])
    return key, value


def _validate_structure(raw_config: Dict[str, Any]) -> ConfigMappingType:
    """Helper function to narrow types for use by SQLFluff.

    This is a recursive function on any dict keys found.
    """
    validated_config: ConfigMappingType = {}
    for key, value in raw_config.items():
        if isinstance(value, dict):
            validated_config[key] = _validate_structure(value)
        elif isinstance(value, list):
            # Coerce all list items to strings, to be in line
            # with the behaviour of ini configs.
            validated_config[key] = [str(item) for item in value]
        elif isinstance(value, (str, int, float, bool)) or value is None:
            validated_config[key] = value
        else:  # pragma: no cover
            # Whatever we found, make it into a string.
            # This is very unlikely to happen and is more for completeness.
            validated_config[key] = str(value)
    return validated_config


def load_toml_file_config(filepath: str) -> ConfigMappingType:
    """Read the SQLFluff config section of a pyproject.toml file.

    We don't need to change any key names here, because the root
    section of the toml file format is `tool.sqlfluff.core`.
    """
    with open(filepath, mode="r") as file:
        toml_dict = tomllib.loads(file.read())
    config_dict = _validate_structure(toml_dict.get("tool", {}).get("sqlfluff", {}))

    # NOTE: For the "rules" section of the sqlfluff config,
    # rule names are often qualified with a dot ".". In the
    # toml scenario this can get interpreted as a nested
    # section, and we resolve that edge case here.
    if "rules" not in config_dict:
        # No rules section, so no need to resolve.
        return config_dict

    rules_section = config_dict["rules"]
    assert isinstance(rules_section, dict), (
        "Expected to find section in `rules` section of config, "
        f"but instead found {rules_section}"
    )
    # Condense the rules section.
    config_dict["rules"] = records_to_nested_dict(
        _condense_rule_record(record)
        for record in iter_records_from_nested_dict(rules_section)
    )

    return config_dict



"""Test the KeywordSegment class."""

from sqlfluff.core.parser import KeywordSegment, StringParser
from sqlfluff.core.parser.context import ParseContext
from sqlfluff.core.parser.match_result import MatchResult


def test__parser__core_keyword(raw_segments):
    """Test the Mystical KeywordSegment."""
    # First make a keyword
    FooKeyword = StringParser("foobar", KeywordSegment, type="bar")
    # Check it looks as expected
    assert FooKeyword.template.upper() == "FOOBAR"
    ctx = ParseContext(dialect=None)
    # Match it against a list and check it doesn't match
    assert not FooKeyword.match(raw_segments, 1, parse_context=ctx)
    # Match it against the final element (returns tuple)
    m = FooKeyword.match(raw_segments, 0, parse_context=ctx)
    assert m
    assert m == MatchResult(
        matched_slice=slice(0, 1),
        matched_class=KeywordSegment,
        segment_kwargs={"instance_types": ("bar",)},
    )
    segments = m.apply(raw_segments)
    assert len(segments) == 1
    segment = segments[0]
    assert segment.class_types == {
        "base",
        "word",
        "keyword",
        "raw",
        "bar",
    }



"""Tests for the configuration routines."""

import logging
import os

import pytest

import sqlfluff
from sqlfluff.core import FluffConfig, Linter
from sqlfluff.core.errors import SQLFluffUserError
from sqlfluff.core.templaters import (
    JinjaTemplater,
    PlaceholderTemplater,
    PythonTemplater,
    RawTemplater,
)
from sqlfluff.utils.testing.logging import fluff_log_catcher

config_b = {
    "core": {"rules": "LT03", "dialect": "ansi"},
    "layout": {
        "type": {"comma": {"line_position": "trailing", "spacing_before": "touch"}}
    },
}

config_c = {
    "core": {"rules": "LT03", "dialect": "ansi"},
    # NOTE:
    # - NOT_A_RULE doesn't match anything.
    # - L001 is an alias, but no longer a rule.
    # - layout is a group and but doesn't match any individual rule.
    "rules": {
        "NOT_A_RULE": {"foo": "bar"},
        "L001": {"foo": "bar"},
        "layout": {"foo": "bar"},
    },
}


def test__config__from_strings():
    """Test loading config from multiple strings."""
    strings = [
        "[sqlfluff]\ndialect=mysql\ntesting_val=foobar",
        "[sqlfluff]\ndialect=postgres\ntesting_val2=bar",
        "[sqlfluff]\ndialect=mysql\ntesting_val=foo",
    ]
    cfg = FluffConfig.from_strings(*strings)
    assert cfg.get("dialect") == "mysql"
    assert cfg.get("testing_val2") == "bar"
    assert cfg.get("testing_val") == "foo"


def test__config__nested_config_tests():
    """Test linting with overridden config in nested paths.

    This looks like a linter test but it's actually a config
    test.
    """
    lntr = Linter(
        # Exclude CP02 in overrides (similar to cli --exclude-rules)
        config=FluffConfig(overrides=dict(exclude_rules="CP02", dialect="ansi"))
    )
    lnt = lntr.lint_path("test/fixtures/config/inheritance_b")
    violations = lnt.check_tuples_by_path()
    for k in violations:
        if k.endswith("nested\\example.sql"):
            # CP01 is enabled in the .sqlfluff file and not excluded.
            assert ("CP01", 1, 4) in violations[k]
            # LT02 is enabled in the .sqlfluff file and not excluded.
            assert ("LT02", 1, 1) in violations[k]
            # CP02 is enabled in the .sqlfluff file but excluded by the
            # override above.
            assert "CP02" not in [c[0] for c in violations[k]]
        elif k.endswith("inheritance_b\\example.sql"):
            # CP01 is enabled because while disabled in the tox.ini file,
            # the exclude-rules option is overridden by the override above
            # which effectively sets the exclude to CP02 and in effect
            # re-enables CP01.
            # This may seem counter-intuitive but is in line with current
            # documentation on how to use `rules` and `exclude-rules`.
            # https://docs.sqlfluff.com/en/latest/perma/rule_disabling.html
            assert ("CP01", 1, 4) in violations[k]
            # CP02 is disabled because of the override above.
            assert "CP02" not in [c[0] for c in violations[k]]
            # LT02 is disabled because it is not in the `rules` of tox.ini
            assert "LT02" not in [c[0] for c in violations[k]]


@pytest.mark.parametrize(
    "templater_name,templater_class,raises_error",
    [
        ("raw", RawTemplater, False),
        ("jinja", JinjaTemplater, False),
        ("python", PythonTemplater, False),
        ("placeholder", PlaceholderTemplater, False),
        ("afefhlsakufe", None, True),
        ("", None, True),
    ],
)
def test__config__templater_selection(templater_name, templater_class, raises_error):
    """Test template selection by name."""
    if raises_error:
        with pytest.raises(SQLFluffUserError):
            FluffConfig(overrides={"dialect": "ansi", "templater": templater_name})
    else:
        cfg = FluffConfig(overrides={"dialect": "ansi", "templater": templater_name})
        assert cfg.get_templater().__class__ is templater_class
        assert cfg._configs["core"]["templater_obj"].__class__ is templater_class


def test__config__glob_exclude_config_tests():
    """Test linting with a glob pattern in exclude_rules.

    This looks like a linter test but it's actually a config
    test.
    """
    lntr = Linter(config=FluffConfig.from_path("test/fixtures/config/glob_exclude"))
    lnt = lntr.lint_path("test/fixtures/config/glob_exclude/test.sql")
    violations = lnt.check_tuples_by_path()
    for k in violations:
        assert ("AM04", 12, 1) in violations[k]
        assert "RF02" not in [c[0] for c in violations[k]]
        assert "LT13" not in [c[0] for c in violations[k]]
        assert "AM05" not in [c[0] for c in violations[k]]
        assert "CV06" not in [c[0] for c in violations[k]]


def test__config__glob_include_config_tests():
    """Test linting with a glob pattern in rules.

    This looks like a linter test but it's actually a config
    test.
    """
    lntr = Linter(config=FluffConfig.from_path("test/fixtures/config/glob_include"))
    lnt = lntr.lint_path("test/fixtures/config/glob_include/test.sql")
    violations = lnt.check_tuples_by_path()
    for k in violations:
        assert ("LT13", 1, 1) in violations[k]
        assert ("AM05", 14, 1) in violations[k]
        assert ("CV06", 14, 9) in violations[k]
        assert ("RF02", 12, 8) in violations[k]
        assert "AM04" not in [c[0] for c in violations[k]]


def test__config__rules_set_to_none():
    """Test linting when rules are set to 'None'.

    Ensure that all rules are still run.
    """
    lntr = Linter(
        config=FluffConfig.from_path("test/fixtures/config/rules_set_to_none")
    )
    lnt = lntr.lint_path("test/fixtures/config/rules_set_to_none/test.sql")
    violations = lnt.check_tuples_by_path()
    for k in violations:
        assert ("LT13", 1, 1) in violations[k]
        assert ("AM04", 12, 1) in violations[k]
        assert ("CP01", 12, 10) in violations[k]


def test__config__rules_group_with_exclude():
    """Test linting when a rules group is selected and rules are excluded."""
    lntr = Linter(
        config=FluffConfig.from_path("test/fixtures/config/rules_group_with_exclude")
    )
    lnt = lntr.lint_path("test/fixtures/config/rules_group_with_exclude/test.sql")
    violations = lnt.check_tuples_by_path()
    for k in violations:
        assert ("CP01", 15, 1) in violations[k]
        assert "LT04" not in [c[0] for c in violations[k]]


def test__config__get_section():
    """Test FluffConfig.get_section method."""
    cfg = FluffConfig(config_b)

    assert cfg.get_section("core").get("rules", None) == "LT03"
    assert cfg.get_section(["layout", "type", "comma"]) == {
        "line_position": "trailing",
        "spacing_before": "touch",
    }
    assert cfg.get_section("non_existent") is None


def test__config__get():
    """Test FluffConfig.get method."""
    cfg = FluffConfig(config_b)

    assert cfg.get("rules") == "LT03"
    assert cfg.get("rulez") is None
    assert cfg.get("rulez", section="core", default=123) == 123
    assert (
        cfg.get("line_position", section=["layout", "type", "comma"], default=None)
        == "trailing"
    )
    assert (
        cfg.get("line_position", section=["layout", "type", "ASDFSDG007"], default=None)
        is None
    )


def test__config__from_kwargs():
    """Test from_kwargs method of FluffConfig."""
    # Instantiate config object.
    cfg = FluffConfig.from_kwargs(
        dialect="snowflake",
        rules=["LT01", "LT02"],
        exclude_rules=["CP01", "AL01"],
    )

    # Verify we can later retrieve the config values.
    assert cfg.get("dialect") == "snowflake"
    assert cfg.get("rules") == "LT01,LT02"
    assert cfg.get("exclude_rules") == "CP01,AL01"


def test__config__from_string():
    """Test from_string method of FluffConfig."""
    with open(
        os.path.join("test", "fixtures", "config", "inheritance_a", ".sqlfluff")
    ) as f:
        config_string = f.read()
    cfg = FluffConfig.from_string(config_string)
    # Verify we can later retrieve the config values.
    assert cfg.get("testing_val") == "foobar"
    assert cfg.get("dialect") == "mysql"


def test__config_missing_dialect():
    """Verify an exception is thrown if no dialect was specified."""
    with pytest.raises(SQLFluffUserError) as e:
        FluffConfig.from_kwargs()
    assert "must configure a dialect" in str(e.value)


def test__config__validate_configs_indirect():
    """Test _validate_configs method of FluffConfig indirectly."""
    # Instantiate config object.
    with pytest.raises(SQLFluffUserError):
        FluffConfig(
            configs={
                "core": {"dialect": "ansi"},
                # This is a known removed value.
                "rules": {"L003": {"lint_templated_tokens": True}},
            }
        )


@pytest.mark.parametrize(
    "raw_sql",
    [
        (
            # "types" not "type"
            "-- sqlfluff:layout:types:comma:line_position:leading\nSELECT 1"
        ),
        (
            # Unsupported layout config length
            "-- sqlfluff:layout:foo:bar\nSELECT 1"
        ),
        (
            # Unsupported layout config length
            "-- sqlfluff:layout:type:comma:bar\nSELECT 1"
        ),
        (
            # Unsupported layout config key ("foo")
            "-- sqlfluff:layout:type:comma:foo:bar\nSELECT 1"
        ),
        (
            # Unsupported layout config key ("foo") [no space]
            "--sqlfluff:layout:type:comma:foo:bar\nSELECT 1"
        ),
    ],
)
def test__config__validate_configs_inline_layout(raw_sql):
    """Test _validate_configs method of FluffConfig when used on a file.

    This test covers both the validation of inline config
    directives but also the validation of layout configs.
    """
    # Instantiate config object.
    cfg = FluffConfig(configs={"core": {"dialect": "ansi"}})

    # Try to process an invalid inline config. Make sure we get an error.
    with pytest.raises(SQLFluffUserError):
        cfg.process_raw_file_for_config(raw_sql, "test.sql")


def test__config__warn_unknown_rule():
    """Test warnings when rules are unknown."""
    lntr = Linter(config=FluffConfig(config_c))

    with fluff_log_catcher(logging.WARNING, "sqlfluff.rules") as caplog:
        lntr.get_rulepack()

    # Check we get a warning on the unrecognised rule.
    assert (
        "Rule configuration contain a section for unexpected rule 'NOT_A_RULE'."
    ) in caplog.text
    # Check we get a warning for the deprecated rule.
    assert (
        "Rule configuration contain a section for unexpected rule 'L001'."
    ) in caplog.text
    # Check we get a hint for the matched rule.
    assert "match for rule LT01 with name 'layout.spacing'" in caplog.text
    # Check we get a warning for the group name.
    assert (
        "Rule configuration contain a section for unexpected rule 'layout'."
    ) in caplog.text
    # Check we get a hint for the matched rule group.
    # NOTE: We don't check the set explicitly because we can't assume ordering.
    assert ("The reference was found as a match for multiple rules: {") in caplog.text
    assert ("LT01") in caplog.text
    assert ("LT02") in caplog.text


def test__process_inline_config():
    """Test the processing of inline in-file configuration directives."""
    cfg = FluffConfig(config_b)
    assert cfg.get("rules") == "LT03"

    cfg.process_inline_config("-- sqlfluff:rules:LT02", "test.sql")
    assert cfg.get("rules") == "LT02"

    assert cfg.get("tab_space_size", section="indentation") == 4
    cfg.process_inline_config("-- sqlfluff:indentation:tab_space_size:20", "test.sql")
    assert cfg.get("tab_space_size", section="indentation") == 20

    assert cfg.get("dialect") == "ansi"
    assert cfg.get("dialect_obj").name == "ansi"
    cfg.process_inline_config("-- sqlfluff:dialect:postgres", "test.sql")
    assert cfg.get("dialect") == "postgres"
    assert cfg.get("dialect_obj").name == "postgres"

    assert cfg.get("rulez") is None
    cfg.process_inline_config("-- sqlfluff:rulez:LT06", "test.sql")
    assert cfg.get("rulez") == "LT06"

    # Check that Windows paths don't get mangled
    cfg.process_inline_config("-- sqlfluff:jinja:my_path:c:\\foo", "test.sql")
    assert cfg.get("my_path", section="jinja") == "c:\\foo"

    # Check that JSON objects are not mangled
    cfg.process_inline_config('-- sqlfluff:jinja:my_dict:{"k":"v"}', "test.sql")
    assert cfg.get("my_dict", section="jinja") == '{"k":"v"}'

    # Check that JSON arrays are not mangled
    cfg.process_inline_config('-- sqlfluff:jinja:my_dict:[{"k":"v"}]', "test.sql")
    assert cfg.get("my_dict", section="jinja") == '[{"k":"v"}]'


@pytest.mark.parametrize(
    "raw_sql",
    [
        (
            "-- sqlfluff:max_line_length:25\n"
            "-- sqlfluff:rules:LT05,LT06\n"
            "-- sqlfluff:exclude_rules:LT01,LT02\n"
            "SELECT 1"
        )
    ],
)
def test__process_raw_file_for_config(raw_sql):
    """Test the processing of a file inline directives."""
    cfg = FluffConfig(config_b)

    # verify initial attributes based on the preloaded configuration
    assert cfg.get("max_line_length") == 80
    assert cfg.get("rules") == "LT03"
    assert cfg.get("exclude_rules") is None

    # internal list attributes should have corresponding exploded list values
    assert cfg.get("rule_allowlist") == ["LT03"]
    assert cfg.get("rule_denylist") == []

    cfg.process_raw_file_for_config(raw_sql, "test.sql")

    # verify overrides based on the file inline directives
    assert cfg.get("max_line_length") == 25
    assert cfg.get("rules") == "LT05,LT06"
    assert cfg.get("exclude_rules") == "LT01,LT02"

    # internal list attributes should have overridden exploded list values
    assert cfg.get("rule_allowlist") == ["LT05", "LT06"]
    assert cfg.get("rule_denylist") == ["LT01", "LT02"]


def test__api__immutable_config():
    """Tests that a config is not mutated when parsing."""
    config = FluffConfig.from_path(
        "test/fixtures/api/config_path_test/extra_configs/.sqlfluff"
    )
    assert config.get("dialect") == "ansi"
    sqlfluff.parse(
        "-- sqlfluff:dialect: postgres\nSELECT * FROM table1\n", config=config
    )
    assert config.get("dialect") == "ansi"



"""Test the Ref grammar.

NOTE: All of these tests depend somewhat on the KeywordSegment working as planned.
"""

import pytest

from sqlfluff.core.dialects import Dialect
from sqlfluff.core.parser.context import ParseContext
from sqlfluff.core.parser.grammar import Ref
from sqlfluff.core.parser.lexer import RegexLexer
from sqlfluff.core.parser.match_result import MatchResult
from sqlfluff.core.parser.parsers import StringParser
from sqlfluff.core.parser.segments import CodeSegment, WhitespaceSegment


@pytest.fixture(scope="function")
def test_dialect():
    """A stripped back test dialect for testing."""
    test_dialect = Dialect("test", root_segment_name="FileSegment")
    test_dialect.set_lexer_matchers(
        [
            RegexLexer("whitespace", r"[^\S\r\n]+", WhitespaceSegment),
            RegexLexer(
                "code", r"[0-9a-zA-Z_]+", CodeSegment, segment_kwargs={"type": "code"}
            ),
        ]
    )
    test_dialect.add(FooSegment=StringParser("foo", CodeSegment, type="foo"))
    # Return the expanded copy.
    return test_dialect.expand()


def test__parser__grammar__ref_eq():
    """Test equality of Ref Grammars."""
    r1 = Ref("foo")
    r2 = Ref("foo")
    assert r1 is not r2
    assert r1 == r2
    check_list = [1, 2, r2, 3]
    # Check we can find it in lists
    assert r1 in check_list
    # Check we can get it's position
    assert check_list.index(r1) == 2
    # Check we can remove it from a list
    check_list.remove(r1)
    assert r1 not in check_list


def test__parser__grammar__ref_repr():
    """Test the __repr__ method of Ref."""
    assert repr(Ref("foo")) == "<Ref: 'foo'>"
    assert repr(Ref("bar", optional=True)) == "<Ref: 'bar' [opt]>"


def test__parser__grammar_ref_match(generate_test_segments, test_dialect):
    """Test the Ref grammar match method."""
    foo_ref = Ref("FooSegment")
    test_segments = generate_test_segments(["bar", "foo", "bar"])
    ctx = ParseContext(dialect=test_dialect)

    match = foo_ref.match(test_segments, 1, ctx)

    assert match == MatchResult(
        matched_slice=slice(1, 2),
        matched_class=CodeSegment,
        segment_kwargs={"instance_types": ("foo",)},
    )


def test__parser__grammar_ref_exclude(generate_test_segments, fresh_ansi_dialect):
    """Test the Ref grammar exclude option with the match method."""
    identifier = Ref("NakedIdentifierSegment", exclude=Ref.keyword("ABS"))
    test_segments = generate_test_segments(["ABS", "ABSOLUTE"])
    ctx = ParseContext(dialect=fresh_ansi_dialect)
    # Assert ABS does not match, due to the exclude
    assert not identifier.match(test_segments, 0, ctx)
    # Assert ABSOLUTE does match
    assert identifier.match(test_segments, 1, ctx)



"""Tests for applying noqa directives and the IgnoreMask."""

import logging
from typing import List

import pytest

from sqlfluff.core import FluffConfig, Linter
from sqlfluff.core.errors import (
    SQLBaseError,
    SQLParseError,
)
from sqlfluff.core.rules.noqa import IgnoreMask, NoQaDirective

# noqa tests require a rule_set, therefore we construct dummy rule set for glob matching.
dummy_rule_map = Linter().get_rulepack().reference_map


class DummyLintError(SQLBaseError):
    """Fake lint error used by tests, similar to SQLLintError."""

    def __init__(self, line_no: int, code: str = "LT01"):
        self._code = code
        super().__init__(line_no=line_no)


def test__linter__raises_malformed_noqa():
    """A badly formatted noqa gets raised as a parsing error."""
    lntr = Linter(dialect="ansi")
    result = lntr.lint_string_wrapped("select 1 --noqa missing semicolon")

    with pytest.raises(SQLParseError):
        result.check_tuples()


@pytest.mark.parametrize(
    "input,expected",
    [
        ("", None),
        ("noqa", NoQaDirective(0, 0, None, None, "noqa")),
        ("noqa?", SQLParseError),
        ("noqa:", NoQaDirective(0, 0, None, None, "noqa:")),
        (
            "noqa:LT01,LT02",
            NoQaDirective(0, 0, ("LT01", "LT02"), None, "noqa:LT01,LT02"),
        ),
        (
            "noqa: enable=LT01",
            NoQaDirective(0, 0, ("LT01",), "enable", "noqa: enable=LT01"),
        ),
        (
            "noqa: disable=CP01",
            NoQaDirective(0, 0, ("CP01",), "disable", "noqa: disable=CP01"),
        ),
        (
            "noqa: disable=all",
            NoQaDirective(0, 0, None, "disable", "noqa: disable=all"),
        ),
        ("noqa: disable", SQLParseError),
        (
            "Inline comment before inline ignore -- noqa:LT01,LT02",
            NoQaDirective(0, 0, ("LT01", "LT02"), None, "noqa:LT01,LT02"),
        ),
        # Test selection with rule globs
        (
            "noqa:L04*",
            NoQaDirective(
                0,
                0,
                (
                    "AM04",  # L044 is an alias of AM04
                    "CP04",  # L040 is an alias of CP04
                    "CV04",  # L047 is an alias of CV04
                    "CV05",  # L049 is an alias of CV05
                    "JJ01",  # L046 is an alias of JJ01
                    "LT01",  # L048 is an alias of LT01
                    "LT10",  # L041 is an alias of LT10
                    "ST02",  # L043 is an alias of ST02
                    "ST03",  # L045 is an alias of ST03
                    "ST05",  # L042 is an alias of ST05
                ),
                None,
                "noqa:L04*",
            ),
        ),
        # Test selection with aliases.
        (
            "noqa:L002",
            NoQaDirective(0, 0, ("LT02",), None, "noqa:L002"),
        ),
        # Test selection with alias globs.
        (
            "noqa:L00*",
            NoQaDirective(
                0,
                0,
                ("LT01", "LT02", "LT03", "LT12"),
                None,
                "noqa:L00*",
            ),
        ),
        # Test selection with names.
        (
            "noqa:capitalisation.keywords",
            NoQaDirective(0, 0, ("CP01",), None, "noqa:capitalisation.keywords"),
        ),
        # Test selection with groups.
        (
            "noqa:capitalisation",
            NoQaDirective(
                0,
                0,
                ("CP01", "CP02", "CP03", "CP04", "CP05"),
                None,
                "noqa:capitalisation",
            ),
        ),
    ],
)
def test_parse_noqa(input, expected):
    """Test correct of "noqa" comments."""
    result = IgnoreMask._parse_noqa(input, 0, 0, reference_map=dummy_rule_map)
    if not isinstance(expected, type):
        assert result == expected
    else:
        # With exceptions, just check the type, not the contents.
        assert isinstance(result, expected)


def test_parse_noqa_no_dups():
    """Test overlapping glob expansions don't return duplicate rules in noqa."""
    result = IgnoreMask._parse_noqa(
        comment="noqa:L0*5,L01*", line_no=0, line_pos=0, reference_map=dummy_rule_map
    )
    assert len(result.rules) == len(set(result.rules))


@pytest.mark.parametrize(
    "noqa,violations,expected,used_noqas",
    [
        [
            [],
            [DummyLintError(1)],
            [
                0,
            ],
            [],
        ],
        [
            [dict(comment="noqa: LT01", line_no=1)],
            [DummyLintError(1)],
            [],
            [0],
        ],
        [
            [dict(comment="noqa: LT01", line_no=2)],
            [DummyLintError(1)],
            [0],
            [],
        ],
        [
            [dict(comment="noqa: LT02", line_no=1)],
            [DummyLintError(1)],
            [0],
            [],
        ],
        [
            [dict(comment="noqa: enable=LT01", line_no=1)],
            [DummyLintError(1)],
            [0],
            [],
        ],
        [
            [dict(comment="noqa: disable=LT01", line_no=1)],
            [DummyLintError(1)],
            [],
            [0],
        ],
        [
            [
                dict(comment="noqa: disable=LT01", line_no=2),
                dict(comment="noqa: enable=LT01", line_no=4),
            ],
            [DummyLintError(1)],
            [0],
            [],  # The disable wasn't used, neither was the enable.
        ],
        [
            [
                dict(comment="noqa: disable=LT01", line_no=2),
                dict(comment="noqa: enable=LT01", line_no=4),
            ],
            [DummyLintError(2)],
            [],
            [0, 1],  # Both were used.
        ],
        [
            [
                dict(comment="noqa: disable=LT01", line_no=2),
                dict(comment="noqa: enable=LT01", line_no=4),
            ],
            [DummyLintError(3)],
            [],
            [0, 1],  # Both were used.
        ],
        [
            [
                dict(comment="noqa: disable=LT01", line_no=2),
                dict(comment="noqa: enable=LT01", line_no=4),
            ],
            [DummyLintError(4)],
            [0],
            [1],  # The enable was matched, but the disable wasn't used.
        ],
        [
            [
                dict(comment="noqa: disable=all", line_no=2),
                dict(comment="noqa: enable=all", line_no=4),
            ],
            [DummyLintError(1)],
            [0],
            # TODO: This is an odd edge case, where we drop out in our
            # evaluation too early so see whether the "enable" is ever
            # matched. In this case _both_ are effectively unused, because
            # we never evaluate the last one. For a first pass I think this
            # might be an acceptable edge case.
            [],
        ],
        [
            [
                dict(comment="noqa: disable=all", line_no=2),
                dict(comment="noqa: enable=all", line_no=4),
            ],
            [DummyLintError(2)],
            [],
            [0, 1],  # Both were used.
        ],
        [
            [
                dict(comment="noqa: disable=all", line_no=2),
                dict(comment="noqa: enable=all", line_no=4),
            ],
            [DummyLintError(3)],
            [],
            [0, 1],  # Both were used.
        ],
        [
            [
                dict(comment="noqa: disable=all", line_no=2),
                dict(comment="noqa: enable=all", line_no=4),
            ],
            [DummyLintError(4)],
            [0],
            [1],  # The enable was matched, but the disable wasn't used.
        ],
        [
            [
                dict(comment="noqa: disable=LT01", line_no=2),
                dict(comment="noqa: enable=all", line_no=4),
            ],
            [
                DummyLintError(2, code="LT01"),
                DummyLintError(2, code="LT02"),
                DummyLintError(4, code="LT01"),
                DummyLintError(4, code="LT02"),
            ],
            [1, 2, 3],
            [0, 1],  # The enable matched. The disable also matched rules.
        ],
        [
            [
                dict(comment="noqa: disable=all", line_no=2),
                dict(comment="noqa: enable=LT01", line_no=4),
            ],
            [
                DummyLintError(2, code="LT01"),
                DummyLintError(2, code="LT02"),
                DummyLintError(4, code="LT01"),
                DummyLintError(4, code="LT02"),
            ],
            [2],
            [0, 1],  # The enable matched the disable. The disable also matched
        ],
        [
            [
                dict(
                    comment="Inline comment before inline ignore -- noqa: LT02",
                    line_no=1,
                )
            ],
            [DummyLintError(1)],
            [0],
            [],
        ],
        [
            [
                dict(
                    comment="Inline comment before inline ignore -- noqa: LT02",
                    line_no=1,
                ),
                dict(
                    comment="Inline comment before inline ignore -- noqa: LT02",
                    line_no=2,
                ),
            ],
            [
                DummyLintError(1),
                DummyLintError(2),
            ],
            [0, 1],
            [],  # Neither used because wrong code.
        ],
        [
            [
                dict(
                    comment="Inline comment before inline ignore -- noqa: L01*",
                    line_no=1,
                ),
            ],
            [
                DummyLintError(1),
            ],
            [0],
            [],  # Neither used because wrong code.
        ],
        [
            [
                dict(
                    comment="Inline comment before inline ignore -- noqa: LT*",
                    line_no=1,
                ),
            ],
            [
                DummyLintError(1),
            ],
            [],
            [0],  # Matched indirectly
        ],
    ],
    ids=[
        "1_violation_no_ignore",
        "1_violation_ignore_specific_line",
        "1_violation_ignore_different_specific_line",
        "1_violation_ignore_different_specific_rule",
        "1_violation_ignore_enable_this_range",
        "1_violation_ignore_disable_this_range",
        "1_violation_line_1_ignore_disable_specific_2_3",
        "1_violation_line_2_ignore_disable_specific_2_3",
        "1_violation_line_3_ignore_disable_specific_2_3",
        "1_violation_line_4_ignore_disable_specific_2_3",
        "1_violation_line_1_ignore_disable_all_2_3",
        "1_violation_line_2_ignore_disable_all_2_3",
        "1_violation_line_3_ignore_disable_all_2_3",
        "1_violation_line_4_ignore_disable_all_2_3",
        "4_violations_two_types_disable_specific_enable_all",
        "4_violations_two_types_disable_all_enable_specific",
        "1_violations_comment_inline_ignore",
        "2_violations_comment_inline_ignore",
        "1_violations_comment_inline_glob_ignore_unmatch",
        "1_violations_comment_inline_glob_ignore_match",
    ],
)
def test_linted_file_ignore_masked_violations(
    noqa: dict, violations: List[SQLBaseError], expected, used_noqas
):
    """Test that _ignore_masked_violations() correctly filters violations."""
    ignore_mask = [
        IgnoreMask._parse_noqa(reference_map=dummy_rule_map, line_pos=0, **c)
        for c in noqa
    ]
    result = IgnoreMask(ignore_mask).ignore_masked_violations(violations)
    expected_violations = [v for i, v in enumerate(violations) if i in expected]
    assert expected_violations == result
    # Check whether "used" evaluation works
    expected_used = [ignore_mask[i] for i, _ in enumerate(noqa) if i in used_noqas]
    actually_used = [i for i in ignore_mask if i.used]
    assert actually_used == expected_used


def test_linter_noqa():
    """Test "noqa" feature at the higher "Linter" level."""
    lntr = Linter(
        config=FluffConfig(
            overrides={
                "dialect": "bigquery",  # Use bigquery to allow hash comments.
                "rules": "AL02, LT04",
            }
        )
    )
    sql = """
    SELECT
        col_a a,
        col_b b, --noqa: disable=AL02
        col_c c,
        col_d d, --noqa: enable=AL02
        col_e e,
        col_f f,
        col_g g,  --noqa
        col_h h,
        col_i i, --noqa:AL02
        col_j j,
        col_k k, --noqa:AL03
        col_l l,
        col_m m,
        col_n n, --noqa: disable=all
        col_o o,
        col_p p, --noqa: enable=all
        col_q q, --Inline comment --noqa: AL02
        col_r r, /* Block comment */ --noqa: AL02
        col_s s # hash comment --noqa: AL02
        -- We trigger both AL02 (implicit aliasing)
        -- and LT04 (leading commas) here to
        -- test glob ignoring of multiple rules.
        , col_t t --noqa: L01*
        , col_u u -- Some comment --noqa: L01*
        , col_v v -- We can ignore both AL02 and LT04 -- noqa: L01[29]
    FROM foo
        """
    result = lntr.lint_string(sql)
    violations = result.get_violations()
    assert {3, 6, 7, 8, 10, 12, 13, 14, 15, 18} == {v.line_no for v in violations}


def test_linter_noqa_with_templating():
    """Similar to test_linter_noqa, but uses templating (Jinja)."""
    lntr = Linter(
        config=FluffConfig(
            overrides={
                "dialect": "bigquery",  # Use bigquery to allow hash comments.
                "templater": "jinja",
                "rules": "LT05",
            }
        )
    )
    sql = "\n"
    '"{%- set a_var = ["1", "2"] -%}\n'
    "SELECT\n"
    "  this_is_just_a_very_long_line_for_demonstration_purposes_of_a_bug_involving_"
    "templated_sql_files, --noqa: LT05\n"
    "  this_is_not_so_big a, --Inline comment --noqa: AL02\n"
    "  this_is_not_so_big b, /* Block comment */ --noqa: AL02\n"
    "  this_is_not_so_big c, # hash comment --noqa: AL02\n"
    "  this_is_just_a_very_long_line_for_demonstration_purposes_of_a_bug_involving_"
    "templated_sql_files, --noqa: L01*\n"
    "FROM\n"
    "  a_table\n"
    "    "
    result = lntr.lint_string(sql)
    assert not result.get_violations()


def test_linter_noqa_template_errors():
    """Similar to test_linter_noqa, but uses templating (Jinja)."""
    lntr = Linter(
        config=FluffConfig(
            overrides={
                "templater": "jinja",
                "dialect": "ansi",
            }
        )
    )
    sql = """select * --noqa: TMP
from raw
where
    balance_date >= {{ execution_date - macros.timedelta() }}  --noqa: TMP
"""
    result = lntr.lint_string(sql)
    assert not result.get_violations()


@pytest.mark.parametrize("disable_noqa", [True, False])
@pytest.mark.parametrize(
    "sql", ["SELEC * FROM foo -- noqa: PRS\n", "{% if 1 > '2' %} -- noqa: TMP\n"]
)
def test_linter_noqa_prs(sql, disable_noqa, caplog):
    """Test "noqa" feature to ignore PRS or TMP at the higher "Linter" level.

    Because templating and parsing failures prevent a fully formed parse tree
    to be formed the rely on slightly different routines to ensure ignores are
    still applied.
    """
    lntr = Linter(
        config=FluffConfig(
            overrides={
                "disable_noqa": disable_noqa,
                "dialect": "ansi",
            }
        )
    )
    with caplog.at_level(logging.DEBUG, logger="sqlfluff.linter"):
        result = lntr.lint_string(sql)
    violations = result.get_violations()
    # In both the templating fail and parsing fail cases, the failures should be
    # ignored because of the inline ignore, _unless_ `disable_noqa`` is set.
    if disable_noqa:
        assert violations
    else:
        assert not violations


def test_linter_noqa_tmp():
    """Test "noqa" feature to ignore TMP at the higher "Linter" level."""
    lntr = Linter(
        config=FluffConfig(
            overrides={
                "exclude_rules": "LT13",
                "dialect": "ansi",
            }
        )
    )
    sql = """
SELECT {{ col_a }} AS a -- noqa: TMP,PRS
FROM foo;
"""
    result = lntr.lint_string(sql)
    print(result.tree.stringify())
    violations = result.get_violations()
    assert not violations


def test_linter_noqa_disable():
    """Test "noqa" comments can be disabled via the config."""
    lntr_noqa_enabled = Linter(
        config=FluffConfig(
            overrides={
                "rules": "AL02",
                "dialect": "ansi",
            }
        )
    )
    lntr_noqa_disabled = Linter(
        config=FluffConfig(
            overrides={
                "disable_noqa": True,
                "rules": "AL02",
                "dialect": "ansi",
            }
        )
    )
    # This query raises AL02, but it is being suppressed by the inline noqa comment.
    # We can ignore this comment by setting disable_noqa = True in the config
    # or by using the --disable-noqa flag in the CLI.
    sql = """
    SELECT col_a a --noqa: AL02
    FROM foo
    """

    # Verify that noqa works as expected with disable_noqa = False (default).
    result_noqa_enabled = lntr_noqa_enabled.lint_string(sql)
    violations_noqa_enabled = result_noqa_enabled.get_violations()
    assert len(violations_noqa_enabled) == 0

    # Verify that noqa comment is ignored with disable_noqa = True.
    result_noqa_disabled = lntr_noqa_disabled.lint_string(sql)
    violations_noqa_disabled = result_noqa_disabled.get_violations()
    assert len(violations_noqa_disabled) == 1
    assert violations_noqa_disabled[0].rule.code == "AL02"


def test_linter_disable_noqa_except():
    """Test "noqa" comments can be disabled via the config."""
    lntr_disable_noqa_except_al02 = Linter(
        config=FluffConfig(
            overrides={
                "disable_noqa_except": "AL02",
                "rules": "AL02, CP01",
                "dialect": "ansi",
            }
        )
    )
    lntr_disable_noqa_except_core = Linter(
        config=FluffConfig(
            overrides={
                "disable_noqa_except": "core",
                "rules": "AL02, CP01",
                "dialect": "ansi",
            }
        )
    )
    # This query raises AL02 and CP01 but it is being suppressed by the inline noqa
    # comments. We can partially ignore these comment by setting
    # disable_noqa_except = rule_list in the config or by using the
    # --disable-noqa-except option in the CLI.
    sql = """
    SELECT
    col_a a, --noqa: AL02
    col_b b --noqa: aliasing
    from foo; --noqa: CP01
    """

    # Verify that noqa comment is ignored with
    # disable_noqa_except = AL02 (base rule name).
    result_disable_noqa_except_al02 = lntr_disable_noqa_except_al02.lint_string(sql)
    violations_disable_noqa_except_al02 = (
        result_disable_noqa_except_al02.get_violations()
    )
    assert len(violations_disable_noqa_except_al02) == 1
    assert violations_disable_noqa_except_al02[0].rule.code == "CP01"

    # Verify that noqa works as expected with disable_noqa_except = core (rule alias).
    result_disable_noqa_except_core = lntr_disable_noqa_except_core.lint_string(sql)
    violations_disable_noqa_except_core = (
        result_disable_noqa_except_core.get_violations()
    )
    assert len(violations_disable_noqa_except_core) == 0



#!/usr/bin/env python

"""Utility strings for use during deployment.

NB: This is not part of the core sqlfluff code.
"""


# This contains various utility scripts

import os
import re
import shutil
import time

import click
from fastcore.net import HTTPError
from ghapi.all import GhApi


@click.group()
def cli():
    """Launch the utility cli."""
    pass


@cli.command()
@click.option("--path", default=".test-reports")
def clean_tests(path):
    """Clear up the tests directory.

    NB: Using scripts allows platform independence
    Makes a new one afterward
    """
    try:
        shutil.rmtree(path)
        click.echo(f"Removed {path!r}...")
    # OSError is for python 27
    # in py36 its FileNotFoundError (but that inherits from IOError, which exists in
    # py27)
    except OSError:
        click.echo(f"Directory {path!r} does not exist. Skipping...")

    os.mkdir(path)
    click.echo(f"Created {path!r}")


@cli.command()
@click.argument("new_version_num")
def release(new_version_num):
    """Change version number in the cfg files.

    NOTE: For fine grained personal access tokens, this requires
    _write_ access to the "contents" scope. For dome reason, if you
    only grant the _read_ access, you can't see any *draft* PRs
    which are necessary for this script to run.
    """
    api = GhApi(
        owner=os.environ["GITHUB_REPOSITORY_OWNER"],
        repo="sqlfluff",
        token=os.environ["GITHUB_TOKEN"],
    )
    try:
        releases = api.repos.list_releases(per_page=100)
    except HTTPError as err:
        raise click.UsageError(
            "HTTP Error from GitHub API. Check your credentials.\n"
            "(i.e. GITHUB_REPOSITORY_OWNER & GITHUB_TOKEN)\n"
            f"{err}"
        )

    latest_draft_release = None
    for rel in releases:
        if rel["draft"]:
            latest_draft_release = rel
            break

    if not latest_draft_release:
        raise click.UsageError(
            "No draft release found on GitHub.\n"
            "This could be because the GitHub action which generates it is broken, "
            "but is more likely due to using an API token which only has read-only "
            "access to the `sqlfluff/sqlfluff` repository. This script requires an "
            "API token with `read and write` access to the `contents` scope in "
            "order to be able to view draft releases."
        )

    # Pre-releases are identifiable because they contain letters.
    # https://peps.python.org/pep-0440/
    is_pre_release = any(char.isalpha() for char in new_version_num)
    click.echo(
        f"Preparing for release {new_version_num}. (Pre-release: {is_pre_release})"
    )

    # Linkify the PRs and authors
    draft_body_parts = latest_draft_release["body"].split("\n")
    potential_new_contributors = []
    for i, p in enumerate(draft_body_parts):
        draft_body_parts[i] = re.sub(
            r"\(#([0-9]*)\) @([^ ]*)$",
            r"[#\1](https://github.com/sqlfluff/sqlfluff/pull/\1) [@\2](https://github.com/\2)",  # noqa E501
            p,
        )
        new_contrib_string = re.sub(
            r".*\(#([0-9]*)\) @([^ ]*)$",
            r"* [@\2](https://github.com/\2) made their first contribution in [#\1](https://github.com/sqlfluff/sqlfluff/pull/\1)",  # noqa E501
            p,
        )
        if new_contrib_string.startswith("* "):
            new_contrib_name = re.sub(r"\* \[(.*?)\].*", r"\1", new_contrib_string)
            potential_new_contributors.append(
                {"name": new_contrib_name, "line": new_contrib_string}
            )
    whats_changed_text = "\n".join(draft_body_parts)

    # Find the first commit for each contributor in this release
    potential_new_contributors.reverse()
    seen_contributors = set()
    deduped_potential_new_contributors = []
    for c in potential_new_contributors:
        if c["name"] not in seen_contributors:
            seen_contributors.add(c["name"])
            deduped_potential_new_contributors.append(c)

    click.echo("Updating CHANGELOG.md...")
    input_changelog = open("CHANGELOG.md", encoding="utf8").readlines()
    write_changelog = open("CHANGELOG.md", "w", encoding="utf8")
    for i, line in enumerate(input_changelog):
        write_changelog.write(line)
        if "DO NOT DELETE THIS LINE" in line:
            existing_entry_start = i + 2
            new_heading = f"## [{new_version_num}] - {time.strftime('%Y-%m-%d')}\n"
            # If the release is already in the changelog, update it
            if f"## [{new_version_num}]" in input_changelog[existing_entry_start]:
                click.echo(f"...found existing entry for {new_version_num}")
                # Update the existing heading with the new date.
                input_changelog[existing_entry_start] = new_heading

                # Delete the existing What’s Changed and New Contributors sections
                remaining_changelog = input_changelog[existing_entry_start:]
                existing_whats_changed_start = (
                    next(
                        j
                        for j, line in enumerate(remaining_changelog)
                        if line.startswith("## What’s Changed")
                    )
                    + existing_entry_start
                )
                existing_new_contributors_start = (
                    next(
                        j
                        for j, line in enumerate(remaining_changelog)
                        if line.startswith("## New Contributors")
                    )
                    + existing_entry_start
                )
                existing_new_contributors_length = (
                    next(
                        j
                        for j, line in enumerate(
                            input_changelog[existing_new_contributors_start:]
                        )
                        if line.startswith("## [")
                    )
                    - 1
                )

                del input_changelog[
                    existing_whats_changed_start : existing_new_contributors_start
                    + existing_new_contributors_length
                ]

                # Now that we've cleared the previous sections, we will accurately
                # find if contributors have been previously mentioned in the changelog
                new_contributor_lines = []
                input_changelog_str = "".join(
                    input_changelog[existing_whats_changed_start:]
                )
                for c in deduped_potential_new_contributors:
                    if c["name"] not in input_changelog_str:
                        new_contributor_lines.append(c["line"])
                input_changelog[existing_whats_changed_start] = (
                    whats_changed_text
                    + "\n\n## New Contributors\n"
                    + "\n".join(new_contributor_lines)
                    + "\n\n"
                )

            else:
                click.echo(f"...creating new entry for {new_version_num}")
                write_changelog.write(f"\n{new_heading}\n## Highlights\n\n")
                write_changelog.write(whats_changed_text)
                write_changelog.write("\n## New Contributors\n\n")
                # Ensure contributor names don't appear in input_changelog list
                new_contributor_lines = []
                input_changelog_str = "".join(input_changelog)
                for c in deduped_potential_new_contributors:
                    if c["name"] not in input_changelog_str:
                        new_contributor_lines.append(c["line"])
                write_changelog.write("\n".join(new_contributor_lines))
                write_changelog.write("\n")

    write_changelog.close()

    click.echo("Updating plugins/sqlfluff-templater-dbt/pyproject.toml")
    for filename in ["plugins/sqlfluff-templater-dbt/pyproject.toml"]:
        input_file = open(filename, "r").readlines()
        # Regardless of platform, write newlines as \n
        write_file = open(filename, "w", newline="\n")
        for line in input_file:
            if line.startswith("version"):
                line = f'version = "{new_version_num}"\n'
            elif line.startswith('    "sqlfluff=='):
                line = f'    "sqlfluff=={new_version_num}",\n'
            write_file.write(line)
        write_file.close()

    keys = ["version"]
    if not is_pre_release:
        # Only update stable_version if it's not a pre-release.
        keys.append("stable_version")

    click.echo("Updating pyproject.toml")
    for filename in ["pyproject.toml"]:
        input_file = open(filename, "r").readlines()
        # Regardless of platform, write newlines as \n
        write_file = open(filename, "w", newline="\n")
        for line in input_file:
            for key in keys:
                if line.startswith(key):
                    # For pyproject.toml we quote the version identifier.
                    line = f'{key} = "{new_version_num}"\n'
                    break
            write_file.write(line)
        write_file.close()

    if not is_pre_release:
        click.echo("Updating gettingstarted.rst")
        for filename in ["docs/source/gettingstarted.rst"]:
            input_file = open(filename, "r").readlines()
            # Regardless of platform, write newlines as \n
            write_file = open(filename, "w", newline="\n")
            change_next_line = False
            for line in input_file:
                if change_next_line:
                    line = f"    {new_version_num}\n"
                    change_next_line = False
                elif line.startswith("    $ sqlfluff version"):
                    change_next_line = True
                write_file.write(line)
            write_file.close()

    click.echo("DONE")


if __name__ == "__main__":
    cli()



"""Test the RawSegment class."""

from sqlfluff.core.parser.segments.base import PathStep


def test__parser__raw_get_raw_segments(raw_segments):
    """Test niche case of calling get_raw_segments on a raw segment."""
    for s in raw_segments:
        assert s.get_raw_segments() == [s]


def test__parser__raw_segments_with_ancestors(
    raw_segments, DummySegment, DummyAuxSegment
):
    """Test raw_segments_with_ancestors.

    This is used in the reflow module to assess parse depth.
    """
    test_seg = DummySegment([DummyAuxSegment(raw_segments[:1]), raw_segments[1]])
    # Result should be the same raw segment, but with appropriate parents
    assert test_seg.raw_segments_with_ancestors == [
        (
            raw_segments[0],
            [
                PathStep(test_seg, 0, 2, (0, 1)),
                PathStep(test_seg.segments[0], 0, 1, (0,)),
            ],
        ),
        (raw_segments[1], [PathStep(test_seg, 1, 2, (0, 1))]),
    ]



"""Tests the python routines within RF02."""

import sqlfluff


def test__rules__std_RF02_wildcard_single_count():
    """Verify that RF02 is only raised once for wildcard (see issue #1973)."""
    sql = """
        SELECT *
        FROM foo
        INNER JOIN bar;
    """
    result = sqlfluff.lint(sql)
    assert "RF02" in [r["code"] for r in result]
    assert [r["code"] for r in result].count("RF02") == 1



"""Tests for the dbt templater."""

import glob
import json
import logging
import os
import pickle
import shutil
import subprocess
from copy import deepcopy
from pathlib import Path
from unittest import mock

import pytest

from sqlfluff.cli.commands import lint
from sqlfluff.core import FluffConfig, Lexer, Linter
from sqlfluff.core.errors import SQLFluffSkipFile, SQLFluffUserError, SQLTemplaterError
from sqlfluff.utils.testing.cli import invoke_assert_code
from sqlfluff.utils.testing.logging import fluff_log_catcher
from sqlfluff_templater_dbt.templater import DbtTemplater


def test__templater_dbt_missing(dbt_templater, project_dir, dbt_fluff_config):
    """Check that a nice error is returned when dbt module is missing."""
    try:
        import dbt  # noqa: F401

        pytest.skip(reason="dbt is installed")
    except ModuleNotFoundError:
        pass

    with pytest.raises(ModuleNotFoundError, match=r"pip install sqlfluff\[dbt\]"):
        dbt_templater.process(
            in_str="",
            fname=os.path.join(project_dir, "models/my_new_project/test.sql"),
            config=FluffConfig(configs=dbt_fluff_config),
        )


def test__templater_dbt_profiles_dir_expanded(dbt_templater):
    """Check that the profiles_dir is expanded."""
    dbt_templater.sqlfluff_config = FluffConfig(
        configs={
            "core": {"dialect": "ansi"},
            "templater": {
                "dbt": {
                    "profiles_dir": "~/.dbt",
                    "profile": "default",
                    "target": "dev",
                    "target_path": "target",
                }
            },
        },
    )
    profiles_dir = dbt_templater._get_profiles_dir()
    # Normalise paths to control for OS variance
    assert os.path.normpath(profiles_dir) == os.path.normpath(
        os.path.expanduser("~/.dbt")
    )
    assert dbt_templater._get_profile() == "default"
    assert dbt_templater._get_target() == "dev"
    assert dbt_templater._get_target_path() == "target"


@pytest.mark.parametrize(
    "fname",
    [
        # dbt_utils
        "use_dbt_utils.sql",
        # macro calling another macro
        "macro_in_macro.sql",
        # config.get(...)
        "use_headers.sql",
        # var(...)
        "use_var.sql",
        # {# {{ 1 + 2 }} #}
        "templated_inside_comment.sql",
        # {{ dbt_utils.last_day(
        "last_day.sql",
        # Many newlines at end, tests templater newline handling
        "trailing_newlines.sql",
        # Ends with whitespace stripping, so trailing newline handling should
        # be disabled
        "ends_with_whitespace_stripping.sql",
        # Access dbt graph nodes
        "access_graph_nodes.sql",
        # Call statements
        "call_statement.sql",
    ],
)
def test__templater_dbt_templating_result(
    project_dir,
    dbt_templater,
    fname,
    dbt_fluff_config,
    dbt_project_folder,
):
    """Test that input sql file gets templated into output sql file."""
    _run_templater_and_verify_result(
        dbt_templater,
        project_dir,
        fname,
        dbt_fluff_config,
        dbt_project_folder,
    )


def test_dbt_profiles_dir_env_var_uppercase(
    project_dir,
    dbt_templater,
    tmpdir,
    monkeypatch,
    dbt_fluff_config,
    dbt_project_folder,
    profiles_dir,
):
    """Tests specifying the dbt profile dir with env var."""
    sub_profiles_dir = tmpdir.mkdir("SUBDIR")  # Use uppercase to test issue 2253
    monkeypatch.setenv("DBT_PROFILES_DIR", str(sub_profiles_dir))
    shutil.copy(os.path.join(profiles_dir, "profiles.yml"), str(sub_profiles_dir))
    _run_templater_and_verify_result(
        dbt_templater,
        project_dir,
        "use_dbt_utils.sql",
        dbt_fluff_config,
        dbt_project_folder,
    )


def _run_templater_and_verify_result(
    dbt_templater,
    project_dir,
    fname,
    dbt_fluff_config,
    dbt_project_folder,
):
    path = Path(project_dir) / "models/my_new_project" / fname
    config = FluffConfig(configs=dbt_fluff_config)
    templated_file, _ = dbt_templater.process(
        in_str=path.read_text(),
        fname=str(path),
        config=config,
    )
    template_output_folder_path = dbt_project_folder / "templated_output/"
    fixture_path = _get_fixture_path(template_output_folder_path, fname)
    assert str(templated_file) == fixture_path.read_text()
    # Check we can lex the output too.
    # https://github.com/sqlfluff/sqlfluff/issues/4013
    lexer = Lexer(config=config)
    _, lexing_violations = lexer.lex(templated_file)
    assert not lexing_violations


def _get_fixture_path(template_output_folder_path, fname):
    fixture_path: Path = template_output_folder_path / fname  # Default fixture location
    dbt_version_specific_fixture_folder = "dbt_utils_0.8.0"
    # Determine where it would exist.
    version_specific_path = (
        Path(template_output_folder_path) / dbt_version_specific_fixture_folder / fname
    )
    if version_specific_path.is_file():
        # Ok, it exists. Use this path instead.
        fixture_path = version_specific_path
    return fixture_path


@pytest.mark.parametrize(
    "fnames_input, fnames_expected_sequence",
    [
        [
            (
                Path("models") / "depends_on_ephemeral" / "a.sql",
                Path("models") / "depends_on_ephemeral" / "b.sql",
                Path("models") / "depends_on_ephemeral" / "d.sql",
            ),
            # c.sql is not present in the original list and should not appear here,
            # even though b.sql depends on it. This test ensures that "out of scope"
            # files, e.g. those ignored using ".sqlfluffignore" or in directories
            # outside what was specified, are not inadvertently processed.
            (
                Path("models") / "depends_on_ephemeral" / "a.sql",
                Path("models") / "depends_on_ephemeral" / "b.sql",
                Path("models") / "depends_on_ephemeral" / "d.sql",
            ),
        ],
        [
            (
                Path("models") / "depends_on_ephemeral" / "a.sql",
                Path("models") / "depends_on_ephemeral" / "b.sql",
                Path("models") / "depends_on_ephemeral" / "c.sql",
                Path("models") / "depends_on_ephemeral" / "d.sql",
            ),
            # c.sql should come before b.sql because b.sql depends on c.sql.
            # It also comes first overall because ephemeral models come first.
            (
                Path("models") / "depends_on_ephemeral" / "c.sql",
                Path("models") / "depends_on_ephemeral" / "a.sql",
                Path("models") / "depends_on_ephemeral" / "b.sql",
                Path("models") / "depends_on_ephemeral" / "d.sql",
            ),
        ],
    ],
)
def test__templater_dbt_sequence_files_ephemeral_dependency(
    project_dir,
    dbt_templater,
    fnames_input,
    fnames_expected_sequence,
    dbt_fluff_config,
):
    """Test that dbt templater sequences files based on dependencies."""
    result = dbt_templater.sequence_files(
        [str(Path(project_dir) / fn) for fn in fnames_input],
        config=FluffConfig(configs=dbt_fluff_config),
    )
    pd = Path(project_dir)
    expected = [str(pd / fn) for fn in fnames_expected_sequence]
    assert list(result) == expected


@pytest.mark.parametrize(
    "raw_file,templated_file,result",
    [
        (
            "select * from a",
            """
with dbt__CTE__INTERNAL_test as (
select * from a
)select count(*) from dbt__CTE__INTERNAL_test
""",
            # The unwrapper should trim the ends.
            [
                ("literal", slice(0, 15, None), slice(0, 15, None)),
            ],
        )
    ],
)
def test__templater_dbt_slice_file_wrapped_test(
    raw_file,
    templated_file,
    result,
    dbt_templater,
    caplog,
):
    """Test that wrapped queries are sliced safely using _check_for_wrapped()."""

    def _render_func(in_str) -> str:
        """Create a dummy render func.

        Importantly one that does actually allow different content to be added.
        """
        # Find the raw location in the template for the test case.
        loc = templated_file.find(raw_file)
        # Replace the new content at the previous position.
        # NOTE: Doing this allows the tracer logic to do what it needs to do.
        return templated_file[:loc] + in_str + templated_file[loc + len(raw_file) :]

    with caplog.at_level(logging.DEBUG, logger="sqlfluff.templater"):
        _, resp, _ = dbt_templater.slice_file(
            raw_file,
            render_func=_render_func,
        )
    assert resp == result


@pytest.mark.parametrize(
    "fname",
    [
        "tests/test.sql",
        "models/my_new_project/single_trailing_newline.sql",
        "models/my_new_project/multiple_trailing_newline.sql",
    ],
)
def test__templater_dbt_templating_test_lex(
    project_dir,
    dbt_templater,
    fname,
    dbt_fluff_config,
):
    """Demonstrate the lexer works on both dbt models and dbt tests.

    Handle any number of newlines.
    """
    path = Path(project_dir) / fname
    config = FluffConfig(configs=dbt_fluff_config)
    source_dbt_sql = path.read_text()
    # Count the newlines.
    n_trailing_newlines = len(source_dbt_sql) - len(source_dbt_sql.rstrip("\n"))
    print(
        f"Loaded {path!r} (n_newlines: {n_trailing_newlines}): " f"{source_dbt_sql!r}",
    )

    templated_file, _ = dbt_templater.process(
        in_str=source_dbt_sql,
        fname=str(path),
        config=config,
    )

    lexer = Lexer(config=config)
    # Test that it successfully lexes.
    _, _ = lexer.lex(templated_file)

    assert (
        templated_file.source_str
        == "select a\nfrom table_a" + "\n" * n_trailing_newlines
    )
    assert (
        templated_file.templated_str
        == "select a\nfrom table_a" + "\n" * n_trailing_newlines
    )


@pytest.mark.parametrize(
    "path,reason",
    [
        (
            "models/my_new_project/disabled_model.sql",
            "it is disabled",
        ),
        (
            "macros/echo.sql",
            "it is a macro",
        ),
    ],
)
def test__templater_dbt_skips_file(
    path,
    reason,
    dbt_templater,
    project_dir,
    dbt_fluff_config,
):
    """A disabled dbt model should be skipped."""
    with pytest.raises(SQLFluffSkipFile, match=reason):
        dbt_templater.process(
            in_str="",
            fname=os.path.join(project_dir, path),
            config=FluffConfig(configs=dbt_fluff_config),
        )


def test_dbt_fails_stdin(dbt_templater, dbt_fluff_config):
    """Reading from stdin is not supported with dbt templater."""
    with pytest.raises(SQLFluffUserError):
        dbt_templater.process(
            in_str="",
            fname="stdin",
            config=FluffConfig(configs=dbt_fluff_config),
        )


@pytest.mark.parametrize(
    "fname",
    [
        "use_var.sql",
        "incremental.sql",
        "single_trailing_newline.sql",
        "ST06_test.sql",
    ],
)
def test__dbt_templated_models_do_not_raise_lint_error(
    project_dir,
    fname,
    caplog,
    dbt_fluff_config,
):
    """Test that templated dbt models do not raise a linting error."""
    linter = Linter(config=FluffConfig(configs=dbt_fluff_config))
    # Log rules output.
    with caplog.at_level(logging.DEBUG, logger="sqlfluff.rules"):
        lnt = linter.lint_path(
            path=os.path.join(project_dir, "models/my_new_project/", fname)
        )
    for linted_file in lnt.files:
        # Log the rendered file to facilitate better debugging of the files.
        print(f"## FILE: {linted_file.path}")
        print("\n\n## RENDERED FILE:\n\n")
        print(linted_file.templated_file.templated_str)
        print("\n\n## PARSED TREE:\n\n")
        print(linted_file.tree.stringify())
        print("\n\n## VIOLATIONS:")
        for idx, v in enumerate(linted_file.violations):
            print(f"   {idx}:{v.to_dict()}")

    violations = lnt.check_tuples()
    assert len(violations) == 0


def _clean_path(glob_expression):
    """Clear out files matching the provided glob expression."""
    for fsp in glob.glob(glob_expression):
        os.remove(fsp)


@pytest.mark.parametrize(
    "path", ["models/my_new_project/issue_1608.sql", "snapshots/issue_1771.sql"]
)
def test__dbt_templated_models_fix_does_not_corrupt_file(
    project_dir,
    path,
    caplog,
    dbt_fluff_config,
):
    """Test issues where previously "sqlfluff fix" corrupted the file."""
    test_glob = os.path.join(project_dir, os.path.dirname(path), "*FIXED.sql")
    _clean_path(test_glob)
    lntr = Linter(config=FluffConfig(configs=dbt_fluff_config))
    with caplog.at_level(logging.INFO, logger="sqlfluff.linter"):
        lnt = lntr.lint_path(os.path.join(project_dir, path), fix=True)
    try:
        lnt.persist_changes(fixed_file_suffix="FIXED")
        with open(os.path.join(project_dir, path + ".after")) as f:
            comp_buff = f.read()
        with open(os.path.join(project_dir, path.replace(".sql", "FIXED.sql"))) as f:
            fixed_buff = f.read()
        assert fixed_buff == comp_buff
    finally:
        _clean_path(test_glob)


def test__templater_dbt_templating_absolute_path(
    project_dir,
    dbt_templater,
    dbt_fluff_config,
):
    """Test that absolute path of input path does not cause RuntimeError."""
    try:
        dbt_templater.process(
            in_str="",
            fname=os.path.abspath(
                os.path.join(project_dir, "models/my_new_project/use_var.sql")
            ),
            config=FluffConfig(configs=dbt_fluff_config),
        )
    except Exception as e:
        pytest.fail(f"Unexpected RuntimeError: {e}")


@pytest.mark.parametrize(
    "fname,exception_msg,exception_class",
    [
        (
            "compiler_error.sql",
            "Compilation Error in model compiler_error "
            "(models/my_new_project/compiler_error.sql)\n  "
            "Unexpected end of template. Jinja was looking for the following tags: "
            "'endfor' or 'else'.",
            SQLFluffUserError,
        ),
        (
            "unknown_ref.sql",
            # https://github.com/sqlfluff/sqlfluff/issues/3849
            "Model 'model.my_new_project.unknown_ref' "
            "(models/my_new_project/unknown_ref.sql) depends on a node named "
            "'i_do_not_exist' which was not found",
            SQLFluffUserError,
        ),
        (
            "unknown_macro.sql",
            # https://github.com/sqlfluff/sqlfluff/issues/3849
            "Compilation Error in model unknown_macro "
            "(models/my_new_project/unknown_macro.sql)\n  'invalid_macro' is "
            "undefined. This can happen when calling a macro that does not exist.",
            SQLTemplaterError,
        ),
        (
            "compile_missing_table.sql",
            # In the test suite we don't get a very helpful error message from dbt
            # but in live testing, the inclusion of the triggering error sometimes
            # gives us something much more useful.
            "because dbt raised a fatal exception during compilation",
            SQLFluffSkipFile,
        ),
    ],
)
def test__templater_dbt_handle_exceptions(
    project_dir,
    dbt_templater,
    dbt_fluff_config,
    dbt_project_folder,
    fname,
    exception_msg,
    exception_class,
):
    """Test that exceptions during compilation are returned as violation."""
    from dbt.adapters.factory import get_adapter

    src_fpath = dbt_project_folder / "error_models" / fname
    target_fpath = os.path.abspath(
        os.path.join(project_dir, "models/my_new_project/", fname)
    )
    # We move the file that throws an error in and out of the project directory
    # as dbt throws an error if a node fails to parse while computing the DAG
    shutil.move(src_fpath, target_fpath)
    try:
        with pytest.raises(exception_class) as excinfo:
            dbt_templater.process(
                in_str="",
                fname=target_fpath,
                config=FluffConfig(
                    configs=dbt_fluff_config, overrides={"dialect": "ansi"}
                ),
            )
    finally:
        shutil.move(target_fpath, src_fpath)
        get_adapter(dbt_templater.dbt_config).connections.release()

    # Debug logging.
    print("Raised:", excinfo.value)
    for trace in excinfo.traceback:
        print(trace)

    # NB: Replace slashes to deal with different platform paths being returned.
    if exception_class is SQLTemplaterError:
        _msg = excinfo.value.desc().replace("\\", "/")
    else:
        _msg = str(excinfo.value).replace("\\", "/")
    assert exception_msg in _msg
    # Ensure that there's no context parent exception, because they don't pickle well.
    # https://github.com/sqlfluff/sqlfluff/issues/6037
    # We *should* be stripping any inherited exceptions from anything returned here.
    # Any residual dbt exceptions are a risk for pickling errors.
    assert not excinfo.value.__context__
    assert not excinfo.value.__cause__
    # We also ensure that the exception can be pickled and unpickled safely.
    # Pickling of exceptions happens during parallel operation and so if it can't
    # be done safely then that will cause bugs.
    pickled_exception = pickle.dumps(excinfo.value)
    roundtrip_exception = pickle.loads(pickled_exception)
    assert isinstance(roundtrip_exception, type(excinfo.value))
    assert str(roundtrip_exception) == str(excinfo.value)


@mock.patch("dbt.adapters.postgres.impl.PostgresAdapter.set_relations_cache")
def test__templater_dbt_handle_database_connection_failure(
    set_relations_cache,
    project_dir,
    dbt_templater,
    dbt_fluff_config,
):
    """Test the result of a failed database connection."""
    from dbt.adapters.factory import get_adapter

    try:
        from dbt.adapters.exceptions import (
            FailedToConnectError as DbtFailedToConnectException,
        )
    except ImportError:
        try:
            from dbt.exceptions import (
                FailedToConnectError as DbtFailedToConnectException,
            )
        except ImportError:
            from dbt.exceptions import (
                FailedToConnectException as DbtFailedToConnectException,
            )

    # Clear the adapter cache to force this test to create a new connection.
    DbtTemplater.adapters.clear()

    set_relations_cache.side_effect = DbtFailedToConnectException("dummy error")

    src_fpath = (
        "plugins/sqlfluff-templater-dbt/test/fixtures/dbt/error_models"
        "/exception_connect_database.sql"
    )
    target_fpath = os.path.abspath(
        os.path.join(
            project_dir, "models/my_new_project/exception_connect_database.sql"
        )
    )
    dbt_fluff_config_fail = deepcopy(dbt_fluff_config)
    dbt_fluff_config_fail["templater"]["dbt"][
        "profiles_dir"
    ] = "plugins/sqlfluff-templater-dbt/test/fixtures/dbt/profiles_yml_fail"
    # We move the file that throws an error in and out of the project directory
    # as dbt throws an error if a node fails to parse while computing the DAG
    shutil.move(src_fpath, target_fpath)
    try:
        with pytest.raises(SQLTemplaterError) as excinfo:
            dbt_templater.process(
                in_str="",
                fname=target_fpath,
                config=FluffConfig(configs=dbt_fluff_config),
            )
    finally:
        shutil.move(target_fpath, src_fpath)
        get_adapter(dbt_templater.dbt_config).connections.release()
    # NB: Replace slashes to deal with different platform paths being returned.
    error_message = excinfo.value.desc().replace("\\", "/")
    assert "dbt tried to connect to the database" in error_message


def test__project_dir_does_not_exist_error(dbt_templater):
    """Test an error is logged if the given dbt project directory doesn't exist."""
    dbt_templater.sqlfluff_config = FluffConfig(
        configs={
            "core": {"dialect": "ansi"},
            "templater": {"dbt": {"project_dir": "./non_existing_directory"}},
        }
    )
    with fluff_log_catcher(logging.ERROR, "sqlfluff.templater") as caplog:
        dbt_project_dir = dbt_templater._get_project_dir()
    assert (
        f"dbt_project_dir: {dbt_project_dir} could not be accessed. " "Check it exists."
    ) in caplog.text


@pytest.mark.parametrize(
    ("model_path", "var_value"),
    [
        ("models/vars_from_cli.sql", "expected_value"),
        ("models/vars_from_cli.sql", [1]),
        ("models/vars_from_cli.sql", {"nested": 1}),
    ],
)
def test__context_in_config_is_loaded(
    project_dir,
    dbt_templater,
    model_path,
    var_value,
    dbt_fluff_config,
):
    """Test that variables inside .sqlfluff are passed to dbt."""
    context = {"passed_through_cli": var_value} if var_value else {}

    config_dict = deepcopy(dbt_fluff_config)
    config_dict["templater"]["dbt"]["context"] = context
    config = FluffConfig(config_dict)

    path = Path(project_dir) / model_path

    processed, violations = dbt_templater.process(
        in_str=path.read_text(), fname=str(path), config=config
    )

    assert violations == []
    assert str(var_value) in processed.templated_str


@pytest.mark.parametrize(
    ("model_path", "var_value"),
    [
        ("models/vars_from_env.sql", "expected_value"),
    ],
)
def test__context_in_env_is_loaded(
    project_dir,
    dbt_templater,
    model_path,
    var_value,
    dbt_fluff_config,
):
    """Test that variables inside env are passed to dbt."""
    os.environ["passed_through_env"] = var_value

    config = FluffConfig(dbt_fluff_config)
    path = Path(project_dir) / model_path

    processed, violations = dbt_templater.process(
        in_str=path.read_text(), fname=str(path), config=config
    )

    assert violations == []
    assert str(var_value) in processed.templated_str


def test__dbt_log_supression(dbt_project_folder):
    """Test that when we try and parse in JSON format we get JSON.

    This actually tests that we can successfully suppress unwanted
    logging from dbt.
    """
    oldcwd = os.getcwd()
    try:
        os.chdir(dbt_project_folder)

        cli_options = [
            "--disable-progress-bar",
            "dbt_project/models/my_new_project/operator_errors.sql",
            "-f",
            "json",
        ]

        result = invoke_assert_code(
            ret_code=1,
            args=[
                lint,
                cli_options,
            ],
        )
        # the CliRunner isn't isolated from the dbt plugin loading
        isolated_lint = subprocess.run(
            ["sqlfluff", "lint"] + cli_options, capture_output=True
        )
    finally:
        os.chdir(oldcwd)
    # Check that the full output parses as json
    parsed = json.loads(result.output)
    assert isolated_lint.returncode == 1
    assert b" Registered adapter:" not in isolated_lint.stdout
    assert isinstance(parsed, list)
    assert len(parsed) == 1
    first_file = parsed[0]
    assert isinstance(first_file, dict)
    # NOTE: Path translation for linux/windows.
    assert (
        first_file["filepath"].replace("\\", "/")
        == "dbt_project/models/my_new_project/operator_errors.sql"
    )
    assert len(first_file["violations"]) == 2



"""Test the BaseFileSegment class."""

from sqlfluff.core.parser import BaseFileSegment


def test__parser__base_segments_file(raw_segments):
    """Test BaseFileSegment to behave as expected."""
    base_seg = BaseFileSegment(raw_segments, fname="/some/dir/file.sql")
    assert base_seg.type == "file"
    assert base_seg.file_path == "/some/dir/file.sql"
    assert base_seg.can_start_end_non_code
    assert base_seg.allow_empty



"""The sqlfluff domain for documenting rules."""

from sphinx import addnodes
from sphinx.directives import ObjectDescription
from sphinx.domains import Domain, ObjType
from sphinx.roles import XRefRole
from sphinx.util.nodes import make_refnode


class SQLFluffRule(ObjectDescription):
    """SQLFluff rule directive for sphinx.

    Rule directives can be used as shown below.

    .. code-block:: rst

        .. sqlfluff:rule:: AM01
                           ambiguous.distinct

            Write the documentation for the rule here.

    To cross reference (i.e. refer to) objects defined like this
    both the code and name reference is available:

    .. code-block:: rst

        :sqlfluff:ref:`CP02`
        :sqlfluff:ref:`capitalisation.identifiers`

    """

    def handle_signature(self, sig, signode):
        """Handle the initial signature of the node.

        This formats the header of the section.
        """
        raw_obj_type = "code" if len(sig) == 4 else "rule"
        obj_type = raw_obj_type.capitalize() + " "
        signode += addnodes.desc_type(obj_type, obj_type)
        signode += addnodes.desc_name(sig, sig)

        fullname = obj_type + sig
        signode["type"] = raw_obj_type
        signode["sig"] = sig
        signode["fullname"] = fullname
        return (fullname, raw_obj_type, sig)

    def add_target_and_index(self, name_cls, sig, signode):
        """Hook to add the permalink and index entries."""
        # Add an ID for permalinks
        node_id = "rule" + "-" + sig
        signode["ids"].append(node_id)
        if len(sig) == 4:
            # If it's a code, add support for legacy links too.
            # Both of these formats have been used in the past.
            signode["ids"].append(f"sqlfluff.rules.Rule_{sig}")
            signode["ids"].append(f"sqlfluff.rules.sphinx.Rule_{sig}")
        # Add to domain for xref resolution
        fluff = self.env.get_domain("sqlfluff")
        fluff.add_rule(sig)
        # Add to index
        self.indexnode["entries"].append(("single", sig, node_id, "", None))

    def _object_hierarchy_parts(self, sig_node):
        return ("bundle", "name")

    def _toc_entry_name(self, sig_node) -> str:
        # NOTE: toctree unpacking issues are due to incorrectly
        # setting _toc_parts.
        sig_node["_toc_parts"] = (
            "bundle",
            sig_node["sig"],
        )
        if len(sig_node["sig"]) == 4:
            # It's a code - don't return TOC entry.
            return ""
        else:
            # It's a name
            return sig_node["sig"]


class SQLFluffDomain(Domain):
    """SQLFluff domain."""

    name = "sqlfluff"
    label = "sqlfluff"

    object_types = {
        "rule": ObjType("rule", "rule", "obj"),
    }

    roles = {
        "ref": XRefRole(),
    }

    directives = {
        "rule": SQLFluffRule,
    }

    initial_data = {
        "rules": [],  # object list
    }

    def get_full_qualified_name(self, node):
        """Get the fully qualified name of the rule."""
        return f"rule.{node.arguments[0]}"

    def get_objects(self):
        """Hook to get all the rules."""
        yield from self.data["rules"]

    def resolve_xref(self, env, fromdocname, builder, typ, target, node, contnode):
        """Hook to resolve xrefs.

        References can be made by code or by name, e.g.
        - :sqlfluff:ref:`LT01`
        - :sqlfluff:ref:`layout.spacing`
        """
        match = [
            (docname, anchor)
            for _, sig, _, docname, anchor, _ in self.get_objects()
            if sig == target
        ]

        if len(match) > 0:
            todocname = match[0][0]
            targ = match[0][1]

            return make_refnode(builder, fromdocname, todocname, targ, contnode, targ)
        else:
            print(f"Failed to match xref: {target!r}")
            return None

    def add_rule(self, signature):
        """Add a new recipe to the domain."""
        name = f"rule.{signature}"
        anchor = f"rule-{signature}"

        # name, dispname, type, docname, anchor, priority
        self.data["rules"].append(
            (name, signature, "Rule", self.env.docname, anchor, 0)
        )


def setup(app):
    """Setup the domain."""
    app.add_domain(SQLFluffDomain)



"""Implementation of Rule AL01."""

from typing import Optional, Tuple, cast

from sqlfluff.core.parser import (
    BaseSegment,
    KeywordSegment,
    RawSegment,
)
from sqlfluff.core.rules import BaseRule, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.utils.reflow import ReflowSequence


class Rule_AL01(BaseRule):
    """Implicit/explicit aliasing of table.

    Aliasing of table to follow preference
    (requiring an explicit ``AS`` is the default).

    **Anti-pattern**

    In this example, the alias ``voo`` is implicit.

    .. code-block:: sql

        SELECT
            voo.a
        FROM foo voo

    **Best practice**

    Add ``AS`` to make it explicit.

    .. code-block:: sql

        SELECT
            voo.a
        FROM foo AS voo

    """

    name = "aliasing.table"
    aliases = ("L011",)
    groups: Tuple[str, ...] = ("all", "aliasing")
    config_keywords = ["aliasing"]
    crawl_behaviour = SegmentSeekerCrawler({"alias_expression"}, provide_raw_stack=True)
    is_fix_compatible = True

    _target_parent_types: Tuple[str, ...] = (
        "from_expression_element",
        "merge_statement",
    )

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        """Implicit aliasing of table/column not allowed. Use explicit `AS` clause.

        We look for the alias segment, and then evaluate its parent and whether
        it contains an AS keyword. This is the _eval function for both AL01 and AL02.
        """
        # Config type hints
        self.aliasing: str

        # AL01 is disabled for Oracle, still run for AL02.
        if context.dialect.name == "oracle" and self.name == "aliasing.table":
            return None

        assert context.segment.is_type("alias_expression")
        if context.parent_stack[-1].is_type(*self._target_parent_types):
            # Search for an AS keyword.
            as_keyword: Optional[BaseSegment]
            for as_keyword in context.segment.segments:
                if as_keyword.raw_upper == "AS":
                    break
            else:
                as_keyword = None

            if as_keyword:
                if self.aliasing == "implicit":
                    self.logger.debug("Removing AS keyword and respacing.")
                    return LintResult(
                        anchor=as_keyword,
                        # Generate the fixes to remove and respace accordingly.
                        fixes=ReflowSequence.from_around_target(
                            as_keyword,
                            context.parent_stack[0],
                            config=context.config,
                        )
                        .without(cast(RawSegment, as_keyword))
                        .respace()
                        .get_fixes(),
                    )

            elif self.aliasing != "implicit":
                self.logger.debug("Inserting AS keyword and respacing.")
                for identifier in context.segment.raw_segments:
                    if identifier.is_code:
                        break
                else:  # pragma: no cover
                    raise NotImplementedError(
                        "Failed to find identifier. Raise this as a bug on GitHub."
                    )
                return LintResult(
                    anchor=context.segment,
                    # Work out the insertion and reflow fixes.
                    fixes=ReflowSequence.from_around_target(
                        identifier,
                        context.parent_stack[0],
                        config=context.config,
                        # Only reflow before, otherwise we catch too much.
                        sides="before",
                    )
                    .insert(
                        KeywordSegment("AS"),
                        target=identifier,
                        pos="before",
                    )
                    .respace()
                    .get_fixes(),
                )
        return None



"""Test the string helpers."""

import pytest

from sqlfluff.core.helpers.string import findall, split_comma_separated_string


@pytest.mark.parametrize(
    "mainstr,substr,positions",
    [
        ("", "", []),
        ("a", "a", [0]),
        ("foobar", "o", [1, 2]),
        ("bar bar bar bar", "bar", [0, 4, 8, 12]),
    ],
)
def test__helpers_string__findall(mainstr, substr, positions):
    """Test _findall."""
    assert list(findall(substr, mainstr)) == positions


@pytest.mark.parametrize(
    "raw_str, expected",
    [
        ("AL01,LT08,AL07", ["AL01", "LT08", "AL07"]),
        ("\nAL01,\nLT08,\nAL07,", ["AL01", "LT08", "AL07"]),
        (["AL01", "LT08", "AL07"], ["AL01", "LT08", "AL07"]),
    ],
)
def test__helpers_string__split_comma_separated_string(raw_str, expected):
    """Tests that string and lists are output correctly."""
    assert split_comma_separated_string(raw_str) == expected



"""Tests for templaters."""

import pytest

from sqlfluff.core.templaters import (
    RawTemplater,
    TemplatedFile,
)
from sqlfluff.core.templaters.base import (
    RawFileSlice,
    TemplatedFileSlice,
    iter_indices_of_newlines,
)


@pytest.mark.parametrize(
    "raw_str,positions",
    [
        ("", []),
        ("foo", []),
        ("foo\nbar", [3]),
        ("\nfoo\n\nbar\nfoo\n\nbar\n", [0, 4, 5, 9, 13, 14, 18]),
    ],
)
def test__indices_of_newlines(raw_str, positions):
    """Test iter_indices_of_newlines."""
    assert list(iter_indices_of_newlines(raw_str)) == positions


def test__templater_raw():
    """Test the raw templater."""
    t = RawTemplater()
    instr = "SELECT * FROM {{blah}}"
    outstr, _ = t.process(in_str=instr, fname="test")
    assert instr == str(outstr)


SIMPLE_FILE_KWARGS = {
    "fname": "test.sql",
    "source_str": "01234\n6789{{foo}}fo\nbarss",
    "templated_str": "01234\n6789x\nfo\nbarss",
    "sliced_file": [
        TemplatedFileSlice(*args)
        for args in [
            ("literal", slice(0, 10, None), slice(0, 10, None)),
            ("templated", slice(10, 17, None), slice(10, 12, None)),
            ("literal", slice(17, 25, None), slice(12, 20, None)),
        ]
    ],
    "raw_sliced": [
        RawFileSlice(*args)
        for args in [
            ("x" * 10, "literal", 0),
            ("x" * 7, "templated", 10),
            ("x" * 8, "literal", 17),
        ]
    ],
}

COMPLEX_FILE_KWARGS = {
    "fname": "test.sql",
    "sliced_file": [
        TemplatedFileSlice(*args)
        for args in [
            ("literal", slice(0, 13, None), slice(0, 13, None)),
            ("comment", slice(13, 29, None), slice(13, 13, None)),
            ("literal", slice(29, 44, None), slice(13, 28, None)),
            ("block_start", slice(44, 68, None), slice(28, 28, None)),
            ("literal", slice(68, 81, None), slice(28, 41, None)),
            ("templated", slice(81, 86, None), slice(41, 42, None)),
            ("literal", slice(86, 110, None), slice(42, 66, None)),
            ("templated", slice(68, 86, None), slice(66, 76, None)),
            ("literal", slice(68, 81, None), slice(76, 89, None)),
            ("templated", slice(81, 86, None), slice(89, 90, None)),
            ("literal", slice(86, 110, None), slice(90, 114, None)),  #
            ("templated", slice(68, 86, None), slice(114, 125, None)),
            ("literal", slice(68, 81, None), slice(125, 138, None)),  #
            ("templated", slice(81, 86, None), slice(138, 139, None)),
            ("literal", slice(86, 110, None), slice(139, 163, None)),
            ("templated", slice(110, 123, None), slice(163, 166, None)),
            ("literal", slice(123, 132, None), slice(166, 175, None)),
            ("block_end", slice(132, 144, None), slice(175, 175, None)),
            ("literal", slice(144, 155, None), slice(175, 186, None)),
            ("block_start", slice(155, 179, None), slice(186, 186, None)),
            ("literal", slice(179, 189, None), slice(186, 196, None)),
            ("templated", slice(189, 194, None), slice(196, 197, None)),
            ("literal", slice(194, 203, None), slice(197, 206, None)),
            ("literal", slice(179, 189, None), slice(206, 216, None)),
            ("templated", slice(189, 194, None), slice(216, 217, None)),
            ("literal", slice(194, 203, None), slice(217, 226, None)),
            ("literal", slice(179, 189, None), slice(226, 236, None)),
            ("templated", slice(189, 194, None), slice(236, 237, None)),
            ("literal", slice(194, 203, None), slice(237, 246, None)),
            ("block_end", slice(203, 215, None), slice(246, 246, None)),
            ("literal", slice(215, 230, None), slice(246, 261, None)),
        ]
    ],
    "raw_sliced": [
        RawFileSlice(*args)
        for args in [
            # All contain dummy strings for now.
            ("x" * 13, "literal", 0),
            ("x" * 16, "comment", 13),
            ("x" * 15, "literal", 29),
            ("x" * 24, "block_start", 44),
            ("x" * 13, "literal", 68),
            ("x" * 5, "templated", 81),
            ("x" * 24, "literal", 86),
            ("x" * 13, "templated", 110),
            ("x" * 9, "literal", 123),
            ("x" * 12, "block_end", 132),
            ("x" * 11, "literal", 144),
            ("x" * 24, "block_start", 155),
            ("x" * 10, "literal", 179),
            ("x" * 5, "templated", 189),
            ("x" * 9, "literal", 194),
            ("x" * 12, "block_end", 203),
            ("x" * 15, "literal", 215),
        ]
    ],
}
COMPLEX_FILE_KWARGS["source_str"] = "".join(
    s.raw for s in COMPLEX_FILE_KWARGS["raw_sliced"]
)


@pytest.mark.parametrize(
    "tf_kwargs,in_charpos,out_line_no,out_line_pos",
    [
        # Simple examples
        (
            SIMPLE_FILE_KWARGS,
            0,
            1,
            1,
        ),
        (
            SIMPLE_FILE_KWARGS,
            20,
            3,
            1,
        ),
        (
            SIMPLE_FILE_KWARGS,
            24,
            3,
            5,
        ),
    ],
)
def test__templated_file_get_line_pos_of_char_pos(
    tf_kwargs,
    in_charpos,
    out_line_no,
    out_line_pos,
):
    """Test TemplatedFile.get_line_pos_of_char_pos."""
    file = TemplatedFile(**tf_kwargs)
    res_line_no, res_line_pos = file.get_line_pos_of_char_pos(in_charpos)
    assert res_line_no == out_line_no
    assert res_line_pos == out_line_pos


@pytest.mark.parametrize(
    "templated_position,inclusive,tf_kwargs,sliced_idx_start,sliced_idx_stop",
    [
        (100, True, COMPLEX_FILE_KWARGS, 10, 11),
        (13, True, COMPLEX_FILE_KWARGS, 0, 3),
        (28, True, COMPLEX_FILE_KWARGS, 2, 5),
        # Check end slicing.
        (12, True, SIMPLE_FILE_KWARGS, 1, 3),
        (20, True, SIMPLE_FILE_KWARGS, 2, 3),
        # Check inclusivity
        (13, False, COMPLEX_FILE_KWARGS, 0, 1),
    ],
)
def test__templated_file_find_slice_indices_of_templated_pos(
    templated_position,
    inclusive,
    tf_kwargs,
    sliced_idx_start,
    sliced_idx_stop,
):
    """Test TemplatedFile._find_slice_indices_of_templated_pos."""
    file = TemplatedFile(**tf_kwargs)
    res_start, res_stop = file._find_slice_indices_of_templated_pos(
        templated_position, inclusive=inclusive
    )
    assert res_start == sliced_idx_start
    assert res_stop == sliced_idx_stop


@pytest.mark.parametrize(
    "in_slice,out_slice,is_literal,tf_kwargs",
    [
        # Simple example
        (
            slice(5, 10),
            slice(5, 10),
            True,
            {
                "sliced_file": [
                    TemplatedFileSlice(
                        "literal", slice(0, 20, None), slice(0, 20, None)
                    )
                ],
                "raw_sliced": [RawFileSlice("x" * 20, "literal", 0)],
                "source_str": "x" * 20,
                "fname": "foo.sql",
            },
        ),
        # Trimming the end of a literal (with things that follow).
        (
            slice(10, 13),
            slice(10, 13),
            True,
            COMPLEX_FILE_KWARGS,
        ),
        # Unrealistic, but should still work
        (
            slice(5, 10),
            slice(55, 60),
            True,
            {
                "sliced_file": [
                    TemplatedFileSlice(
                        "literal", slice(50, 70, None), slice(0, 20, None)
                    )
                ],
                "raw_sliced": [
                    RawFileSlice("x" * 50, "literal", 0),
                    RawFileSlice("x" * 20, "literal", 50),
                ],
                "source_str": "x" * 70,
                "fname": "foo.sql",
            },
        ),
        # Spanning a template
        (
            slice(5, 15),
            slice(5, 20),
            False,
            SIMPLE_FILE_KWARGS,
        ),
        # Handling templated
        (
            slice(5, 15),
            slice(0, 25),
            False,
            # NB: Same as SIMPLE_SLICED_FILE, but with different slice types.
            {
                **SIMPLE_FILE_KWARGS,
                "sliced_file": [
                    TemplatedFileSlice(
                        "templated", slc.source_slice, slc.templated_slice
                    )
                    for slc in SIMPLE_FILE_KWARGS["sliced_file"]
                ],
                "raw_sliced": [
                    RawFileSlice(slc.raw, "templated", slc.source_idx)
                    for slc in SIMPLE_FILE_KWARGS["raw_sliced"]
                ],
            },
        ),
        # Handling single length slices
        (
            slice(10, 10),
            slice(10, 10),
            True,
            SIMPLE_FILE_KWARGS,
        ),
        (
            slice(12, 12),
            slice(17, 17),
            True,
            SIMPLE_FILE_KWARGS,
        ),
        # Dealing with single length elements
        (
            slice(20, 20),
            slice(25, 25),
            True,
            {
                "sliced_file": SIMPLE_FILE_KWARGS["sliced_file"]
                + [
                    TemplatedFileSlice(
                        "comment", slice(25, 35, None), slice(20, 20, None)
                    )
                ],
                "raw_sliced": SIMPLE_FILE_KWARGS["raw_sliced"]
                + [RawFileSlice("x" * 10, "comment", 25)],
                "source_str": SIMPLE_FILE_KWARGS["source_str"] + "x" * 10,
                "fname": "foo.sql",
            },
        ),
        # Just more test coverage
        (
            slice(43, 43),
            slice(87, 87),
            True,
            COMPLEX_FILE_KWARGS,
        ),
        (
            slice(13, 13),
            slice(13, 13),
            True,
            COMPLEX_FILE_KWARGS,
        ),
        (
            slice(186, 186),
            slice(155, 155),
            True,
            COMPLEX_FILE_KWARGS,
        ),
        # Backward slicing.
        (
            slice(100, 130),
            # NB This actually would reference the wrong way around if we
            # just take the points. Here we should handle it gracefully.
            slice(68, 110),
            False,
            COMPLEX_FILE_KWARGS,
        ),
    ],
)
def test__templated_file_templated_slice_to_source_slice(
    in_slice, out_slice, is_literal, tf_kwargs
):
    """Test TemplatedFile.templated_slice_to_source_slice."""
    file = TemplatedFile(**tf_kwargs)
    source_slice = file.templated_slice_to_source_slice(in_slice)
    literal_test = file.is_source_slice_literal(source_slice)
    assert (is_literal, source_slice) == (literal_test, out_slice)


@pytest.mark.parametrize(
    "file,expected_result",
    [
        # Comment example
        (
            TemplatedFile(
                source_str=("a" * 10) + "{# b #}" + ("a" * 10),
                fname="test",
                sliced_file=[
                    TemplatedFileSlice("literal", slice(0, 10), slice(0, 10)),
                    TemplatedFileSlice("templated", slice(10, 17), slice(10, 10)),
                    TemplatedFileSlice("literal", slice(17, 27), slice(10, 20)),
                ],
                raw_sliced=[
                    RawFileSlice("a" * 10, "literal", 0),
                    RawFileSlice("{# b #}", "comment", 10),
                    RawFileSlice("a" * 10, "literal", 17),
                ],
            ),
            [RawFileSlice("{# b #}", "comment", 10)],
        ),
        # Template tags aren't source only.
        (
            TemplatedFile(
                source_str=r"aaa{{ b }}aaa",
                fname="test",
                sliced_file=[
                    TemplatedFileSlice("literal", slice(0, 3), slice(0, 3)),
                    TemplatedFileSlice("templated", slice(3, 10), slice(3, 6)),
                    TemplatedFileSlice("literal", slice(10, 13), slice(6, 9)),
                ],
                raw_sliced=[
                    RawFileSlice("aaa", "literal", 0),
                    RawFileSlice("{{ b }}", "templated", 3),
                    RawFileSlice("aaa", "literal", 10),
                ],
            ),
            [],
        ),
    ],
)
def test__templated_file_source_only_slices(file, expected_result):
    """Test TemplatedFile.source_only_slices."""
    assert file.source_only_slices() == expected_result



"""Methods for validating config dicts."""

from sqlfluff.core.config.removed import validate_config_dict_for_removed
from sqlfluff.core.errors import SQLFluffUserError
from sqlfluff.core.types import ConfigMappingType

ALLOWABLE_LAYOUT_CONFIG_KEYS = (
    "spacing_before",
    "spacing_after",
    "spacing_within",
    "line_position",
    "align_within",
    "align_scope",
    "keyword_line_position",
)


def _validate_layout_config(config: ConfigMappingType, logging_reference: str) -> None:
    """Validate the layout config section of the config.

    We check for valid key values and for the depth of the
    structure.

    NOTE: For now we don't check that the "type" is a valid one
    to reference, or that the values are valid. For the values,
    these are likely to be rejected by the layout routines at
    runtime. The last risk area is validating that the type is
    a valid one, but that should be handled by the same as the above.
    """
    layout_section = config.get("layout", {})
    if not layout_section:
        return None

    preamble = f"Config file {logging_reference!r} set an invalid `layout` option. "
    reference = (
        "See https://docs.sqlfluff.com/en/stable/perma/layout.html"
        "#configuring-layout for more details."
    )

    if not isinstance(layout_section, dict):
        raise SQLFluffUserError(
            preamble
            + f"Found value {layout_section!r} instead of a valid layout section. "
            + reference
        )

    # The sections within layout can only be "type" (currently).
    non_type_keys = set(layout_section.keys()) - set(("type",))
    type_section = layout_section.get("type", {})
    if non_type_keys or not type_section or not isinstance(type_section, dict):
        raise SQLFluffUserError(
            preamble
            + "Only sections of the form `sqlfluff:layout:type:...` are valid. "
            + reference
        )

    for layout_type, layout_section in type_section.items():
        if not isinstance(layout_section, dict):
            raise SQLFluffUserError(
                preamble
                + f"Layout config for {layout_type!r} is invalid. Expected a section. "
                + reference
            )

        invalid_keys = set(layout_section.keys()) - set(ALLOWABLE_LAYOUT_CONFIG_KEYS)
        if invalid_keys:
            raise SQLFluffUserError(
                preamble
                + f"Layout config for type {layout_type!r} is invalid. "
                + f"Found the following invalid keys: {invalid_keys}. "
                + reference
            )

        for key in ALLOWABLE_LAYOUT_CONFIG_KEYS:
            if key in layout_section:
                if isinstance(layout_section[key], dict):
                    raise SQLFluffUserError(
                        preamble
                        + f"Layout config for type {layout_type!r} is invalid. "
                        + "Found the an unexpected section rather than "
                        + f"value for {key}. "
                        + reference
                    )


def validate_config_dict(config: ConfigMappingType, logging_reference: str) -> None:
    """Validate a config dict.

    Currently we validate:
    - Removed and deprecated values.
    - Layout configuration structure.

    Using this method ensures that any later validation will also be applied.

    NOTE: Some of these method may mutate the config object where they are
    able to correct issues.
    """
    # Validate the config for any removed values
    validate_config_dict_for_removed(config, logging_reference)
    # Validate layout section
    _validate_layout_config(config, logging_reference)



"""Tests for the Linter class and LintingResult class."""

import logging
import os
from unittest.mock import patch

import pytest

from sqlfluff.cli.formatters import OutputStreamFormatter
from sqlfluff.cli.outputstream import make_output_stream
from sqlfluff.core import FluffConfig, Linter
from sqlfluff.core.errors import (
    SQLBaseError,
    SQLFluffSkipFile,
    SQLLexError,
    SQLLintError,
    SQLParseError,
    SQLTemplaterError,
)
from sqlfluff.core.linter import runner
from sqlfluff.core.linter.linting_result import combine_dicts, sum_dicts
from sqlfluff.core.linter.runner import get_runner
from sqlfluff.utils.testing.logging import fluff_log_catcher


class DummyLintError(SQLBaseError):
    """Fake lint error used by tests, similar to SQLLintError."""

    def __init__(self, line_no: int, code: str = "LT01"):
        self._code = code
        super().__init__(line_no=line_no)


def normalise_paths(paths):
    """Test normalising paths.

    NB Paths on difference platforms might look different, so this
    makes them comparable.
    """
    return {pth.replace("/", ".").replace("\\", ".") for pth in paths}


@pytest.mark.parametrize("filesize,raises_skip", [(0, False), (5, True), (2000, False)])
def test__linter__skip_large_bytes(filesize, raises_skip):
    """Test extracting paths from a file path."""
    config = FluffConfig(
        overrides={"large_file_skip_byte_limit": filesize, "dialect": "ansi"}
    )
    # First check the function directly
    if raises_skip:
        with pytest.raises(SQLFluffSkipFile) as excinfo:
            Linter.load_raw_file_and_config(
                "test/fixtures/linter/indentation_errors.sql", config
            )
        assert "Skipping" in str(excinfo.value)
        assert f"over the limit of {filesize}" in str(excinfo.value)
    # If NOT raises, then we'll catch the raise an error and the test will fail.

    # Then check that it either is or isn't linted appropriately via lint_paths.
    lntr = Linter(config)
    result = lntr.lint_paths(
        ("test/fixtures/linter/indentation_errors.sql",),
    )
    if raises_skip:
        assert not result.get_violations()
    else:
        assert result.get_violations()

    # Same again via parse_path, which is the other entry point.
    result = list(
        lntr.parse_path(
            "test/fixtures/linter/indentation_errors.sql",
        )
    )
    if raises_skip:
        assert not result
    else:
        assert result


@pytest.mark.parametrize(
    "path",
    [
        "test/fixtures/linter/indentation_errors.sql",
        "test/fixtures/linter/whitespace_errors.sql",
    ],
)
def test__linter__lint_string_vs_file(path):
    """Test the linter finds the same things on strings and files."""
    with open(path) as f:
        sql_str = f.read()
    lntr = Linter(dialect="ansi")
    assert (
        lntr.lint_string(sql_str).check_tuples() == lntr.lint_path(path).check_tuples()
    )


@pytest.mark.parametrize(
    "rules,num_violations", [(None, 7), ("CP01", 2), (("LT01", "LT12"), 1)]
)
def test__linter__get_violations_filter_rules(rules, num_violations):
    """Test filtering violations by which rules were violated."""
    lntr = Linter(dialect="ansi")
    lint_result = lntr.lint_string("select a, b FROM tbl c order BY d")

    assert len(lint_result.get_violations(rules=rules)) == num_violations


def test__linter__linting_result__sum_dicts():
    """Test the summing of dictionaries in the linter."""
    i = {}
    a = dict(a=3, b=123, f=876.321)
    b = dict(a=19, b=321.0, g=23478)
    r = dict(a=22, b=444.0, f=876.321, g=23478)
    assert sum_dicts(a, b) == r
    # Check the identity too
    assert sum_dicts(r, i) == r


def test__linter__linting_result__combine_dicts():
    """Test the combination of dictionaries in the linter."""
    a = dict(a=3, b=123, f=876.321)
    b = dict(h=19, i=321.0, j=23478)
    r = dict(z=22)
    assert combine_dicts(a, b, r) == dict(
        a=3, b=123, f=876.321, h=19, i=321.0, j=23478, z=22
    )


def test__linter__linting_result_check_tuples():
    """Test that a LintingResult can partition violations by the source files."""
    lntr = Linter()
    result = lntr.lint_paths(
        (
            "test/fixtures/linter/comma_errors.sql",
            "test/fixtures/linter/whitespace_errors.sql",
        )
    )
    check_tuples = result.check_tuples()
    isinstance(check_tuples, list)
    assert check_tuples == [
        ("LT09", 2, 1),
        ("LT04", 4, 5),
        ("LT02", 5, 1),
        ("LT04", 5, 1),
        ("LT02", 6, 1),
        ("AL02", 6, 5),
        ("LT01", 6, 6),
        ("CP01", 8, 1),
        ("LT09", 1, 1),
        ("LT01", 2, 9),
        ("LT01", 3, 12),
        ("LT02", 4, 1),
        ("CP01", 6, 10),
    ]


def test__linter__linting_result_check_tuples_by_path():
    """Test that a LintingResult can partition violations by the source files."""
    lntr = Linter()
    result = lntr.lint_paths(
        (
            "test/fixtures/linter/comma_errors.sql",
            "test/fixtures/linter/whitespace_errors.sql",
        )
    )
    check_tuples = result.check_tuples_by_path()
    isinstance(check_tuples, dict)
    # Normalise the paths in the keys.
    check_tuples = {k.replace("\\", "/"): v for k, v in check_tuples.items()}
    assert check_tuples == {
        "test/fixtures/linter/comma_errors.sql": [
            ("LT09", 2, 1),
            ("LT04", 4, 5),
            ("LT02", 5, 1),
            ("LT04", 5, 1),
            ("LT02", 6, 1),
            ("AL02", 6, 5),
            ("LT01", 6, 6),
            ("CP01", 8, 1),
        ],
        "test/fixtures/linter/whitespace_errors.sql": [
            ("LT09", 1, 1),
            ("LT01", 2, 9),
            ("LT01", 3, 12),
            ("LT02", 4, 1),
            ("CP01", 6, 10),
        ],
    }


@pytest.mark.parametrize(
    "path,stats",
    [
        (
            "multifile_a",
            {
                "avg per file": 2.5,
                "clean": 0,
                "clean files": 0,
                "exit code": 111,
                "files": 2,
                "status": "FAIL",
                "unclean": 2,
                "unclean files": 2,
                "unclean rate": 1.0,
                "violations": 5,
            },
        ),
        (
            "multifile_b",
            {
                "avg per file": 2.0,
                "clean": 0,
                "clean files": 0,
                "exit code": 111,
                "files": 2,
                "status": "FAIL",
                "unclean": 2,
                "unclean files": 2,
                "unclean rate": 1.0,
                "violations": 4,
            },
        ),
    ],
)
def test__linter__linting_result_stats(path, stats):
    """Test that a LintingResult can get the right stats with multiple files.

    https://github.com/sqlfluff/sqlfluff/issues/5673
    """
    lntr = Linter()
    result = lntr.lint_paths((f"test/fixtures/linter/exit_codes/{path}",))
    # NOTE: We're using fake return codes for testing purposes.
    assert result.stats(111, 222) == stats


@pytest.mark.parametrize("processes", [1, 2])
def test__linter__linting_result_get_violations(processes):
    """Test that we can get violations from a LintingResult."""
    lntr = Linter()
    result = lntr.lint_paths(
        (
            "test/fixtures/linter/comma_errors.sql",
            "test/fixtures/linter/whitespace_errors.sql",
        ),
        processes=processes,
    )

    all([isinstance(v, SQLLintError) for v in result.get_violations()])


@pytest.mark.parametrize("force_error", [False, True])
def test__linter__linting_parallel_thread(force_error, monkeypatch):
    """Run linter in parallel mode using threads.

    Similar to test__linter__linting_result_get_violations but uses a thread
    pool of 1 worker to test parallel mode without subprocesses. This lets the
    tests capture code coverage information for the backend parts of parallel
    execution without having to jump through hoops.
    """
    if not force_error:
        monkeypatch.setattr(Linter, "allow_process_parallelism", False)

    else:

        def _create_pool(*args, **kwargs):
            class ErrorPool:
                def __enter__(self):
                    return self

                def __exit__(self, exc_type, exc_val, exc_tb):
                    pass

                def imap_unordered(self, *args, **kwargs):
                    yield runner.DelayedException(ValueError())

            return ErrorPool()

        monkeypatch.setattr(runner.MultiProcessRunner, "_create_pool", _create_pool)

    config = FluffConfig(overrides={"dialect": "ansi"})
    output_stream = make_output_stream(config, None, os.devnull)
    lntr = Linter(
        formatter=OutputStreamFormatter(output_stream, False, verbosity=0),
        dialect="ansi",
    )
    result = lntr.lint_paths(
        # NOTE: Lint more than one file to make sure we enabled the multithreaded
        # code path.
        (
            "test/fixtures/linter/comma_errors.sql",
            "test/fixtures/linter/whitespace_errors.sql",
        ),
        processes=2,
    )

    all([isinstance(v, SQLLintError) for v in result.get_violations()])


@patch("sqlfluff.core.linter.Linter.lint_rendered")
def test_lint_path_parallel_wrapper_exception(patched_lint):
    """Tests the error catching behavior of _lint_path_parallel_wrapper().

    Test on MultiThread runner because otherwise we have pickling issues.
    """
    patched_lint.side_effect = ValueError("Something unexpected happened")
    for result in runner.MultiThreadRunner(
        Linter(), FluffConfig(overrides={"dialect": "ansi"}), processes=1
    ).run(
        ["test/fixtures/linter/passing.sql"],
        fix=False,
    ):
        assert isinstance(result, runner.DelayedException)
        with pytest.raises(ValueError):
            result.reraise()


@pytest.mark.parametrize(
    "mock_cpu,in_processes,exp_processes",
    [
        # Make the mocked cpu count a really high value which is
        # unlikely to collide with the real value. We can then
        # test all the different combos.
        (512, 1, 1),
        (512, 0, 512),
        (512, -12, 500),
        (512, 5, 5),
        # Check that we can't go lower than 1 in a 1 cpu case
        (1, -1, 1),
    ],
)
@patch("multiprocessing.cpu_count")
def test__linter__get_runner_processes(
    patched_cpu_count, mock_cpu, in_processes, exp_processes
):
    """Test that get_runner handles processes correctly."""
    # Make the mocked cpu count a really high value which is
    # unlikely to collide with the real value.
    patched_cpu_count.return_value = mock_cpu
    _, return_processes = get_runner(
        linter=Linter(),
        config=FluffConfig(overrides={"dialect": "ansi"}),
        processes=in_processes,
    )
    assert return_processes == exp_processes


@patch("sqlfluff.core.linter.runner.linter_logger")
@patch("sqlfluff.core.linter.Linter.lint_rendered")
def test__linter__linting_unexpected_error_handled_gracefully(
    patched_lint, patched_logger
):
    """Test that an unexpected internal error returns the issue-surfacing file."""
    patched_lint.side_effect = Exception("Something unexpected happened")
    lntr = Linter()
    lntr.lint_paths(("test/fixtures/linter/passing.sql",))
    assert (
        "Unable to lint test/fixtures/linter/passing.sql due to an internal error."
        # NB: Replace is to handle windows-style paths.
        in patched_logger.warning.call_args[0][0].replace("\\", "/")
        and "Exception: Something unexpected happened"
        in patched_logger.warning.call_args[0][0]
    )


def test__linter__empty_file():
    """Test linter behaves nicely with an empty string.

    Much of this test is about making sure that ParsedString is
    instantiated appropriately.
    """
    lntr = Linter(dialect="ansi")
    # Make sure no exceptions raised and no violations found in empty file.
    parsed = lntr.parse_string("")
    # There should still be a parsed variant
    assert parsed.parsed_variants
    assert len(parsed.parsed_variants) == 1
    root_variant = parsed.parsed_variants[0]
    # That root variant should still have a templated file and a parsed tree
    # (although that parsed tree will likely just be an end of file marker).
    assert root_variant.templated_file
    assert root_variant.tree
    # No violations
    assert not parsed.violations


def test__linter__parse_fail():
    """Test linter behaves as expected with an unparsable string.

    Much of this test is about making sure that ParsedString is
    instantiated appropriately.
    """
    lntr = Linter(dialect="ansi")
    # Try and parse something which obviously isn't SQL
    parsed = lntr.parse_string("THIS IS NOT SQL")
    # There should still be a parsed variant
    assert parsed.parsed_variants
    assert len(parsed.parsed_variants) == 1
    root_variant = parsed.parsed_variants[0]
    # That root variant should still have a templated file and a parsed tree...
    assert root_variant.templated_file
    assert root_variant.tree
    # ...but that tree should contain an unparsable segment.
    assert "unparsable" in root_variant.tree.type_set()
    # There *should* be violations because there should be a parsing fail.
    assert parsed.violations
    assert any(isinstance(v, SQLParseError) for v in parsed.violations)


def test__linter__templating_fail():
    """Test linter behaves as expected with invalid jinja template.

    Much of this test is about making sure that ParsedString is
    instantiated appropriately.
    """
    lntr = Linter(dialect="ansi")
    # Try and parse something which breaks Jinja templating.
    parsed = lntr.parse_string("{% if foo %}")
    # For a templating fail, there won't be a parsed variant.
    assert not parsed.parsed_variants
    # There *should* be violations because there should be a templating fail.
    assert parsed.violations
    assert any(isinstance(v, SQLTemplaterError) for v in parsed.violations)


@pytest.mark.parametrize(
    "path,rules,ignore_templated_areas,check_tuples",
    [
        (
            "test/fixtures/templater/jinja_h_macros/jinja.sql",
            "L006",
            True,
            [("LT01", 3, 39), ("LT01", 3, 40)],
        ),
        (
            "test/fixtures/templater/jinja_h_macros/jinja.sql",
            "L006",
            False,
            [
                # there are still two of each because LT01 checks
                # for both *before* and *after* the operator.
                # The deduplication filter makes sure there aren't 4.
                ("LT01", 3, 16),
                ("LT01", 3, 16),
                ("LT01", 3, 39),
                ("LT01", 3, 40),
            ],
        ),
        (
            "test/fixtures/linter/jinja_variants/simple_CP01.sql",
            "CP01",
            False,
            [
                # We should get violations from both sides of the if
                # statement without doubling up on the one outside.
                ("CP01", 2, 10),
                ("CP01", 2, 34),
                ("CP01", 2, 52),
            ],
        ),
    ],
)
def test__linter__mask_templated_violations(
    path, rules, ignore_templated_areas, check_tuples
):
    """Test linter masks files properly around templated content.

    NOTE: this also tests deduplication of fixes which have the same
    source position. i.e. `LintedFile.deduplicate_in_source_space()`.
    """
    lntr = Linter(
        config=FluffConfig(
            overrides={
                "rules": rules,
                "ignore_templated_areas": ignore_templated_areas,
                "dialect": "ansi",
            }
        )
    )
    linted = lntr.lint_path(path=path)
    assert linted.check_tuples() == check_tuples


@pytest.mark.parametrize(
    "fname,config_encoding,lexerror",
    [
        (
            "test/fixtures/linter/encoding-utf-8.sql",
            "autodetect",
            False,
        ),
        (
            "test/fixtures/linter/encoding-utf-8-sig.sql",
            "autodetect",
            False,
        ),
        (
            "test/fixtures/linter/encoding-utf-8.sql",
            "utf-8",
            False,
        ),
        (
            "test/fixtures/linter/encoding-utf-8-sig.sql",
            "utf-8",
            True,
        ),
        (
            "test/fixtures/linter/encoding-utf-8.sql",
            "utf-8-sig",
            False,
        ),
        (
            "test/fixtures/linter/encoding-utf-8-sig.sql",
            "utf-8-sig",
            False,
        ),
    ],
)
def test__linter__encoding(fname, config_encoding, lexerror):
    """Test linter deals with files with different encoding."""
    lntr = Linter(
        config=FluffConfig(
            overrides={
                "rules": "LT01",
                "encoding": config_encoding,
                "dialect": "ansi",
            }
        )
    )
    result = lntr.lint_paths((fname,))
    assert lexerror == (SQLLexError in [type(v) for v in result.get_violations()])


def test_delayed_exception():
    """Test that DelayedException stores and reraises a stored exception."""
    ve = ValueError()
    de = runner.DelayedException(ve)
    with pytest.raises(ValueError):
        de.reraise()


def test__attempt_to_change_templater_warning():
    """Test warning when changing templater in .sqlfluff file in subdirectory."""
    initial_config = FluffConfig(
        configs={"core": {"templater": "jinja", "dialect": "ansi"}}
    )
    lntr = Linter(config=initial_config)
    updated_config = FluffConfig(
        configs={"core": {"templater": "python", "dialect": "ansi"}}
    )
    with fluff_log_catcher(logging.WARNING, "sqlfluff.linter") as caplog:
        lntr.render_string(
            in_str="select * from table",
            fname="test.sql",
            config=updated_config,
            encoding="utf-8",
        )
    assert "Attempt to set templater to " in caplog.text


def test_advanced_api_methods():
    """Test advanced API methods on segments."""
    # These aren't used by the simple API, which returns
    # a simple JSON representation of the parse tree, but
    # are available for advanced API usage and within rules.
    sql = """
    WITH cte AS (
        SELECT * FROM tab_a
    )
    SELECT
        cte.col_a,
        tab_b.col_b
    FROM cte
    INNER JOIN tab_b;
    """
    linter = Linter(dialect="ansi")
    parsed = linter.parse_string(sql)

    # CTEDefinitionSegment.get_identifier
    cte_segment = next(parsed.tree.recursive_crawl("common_table_expression"))
    assert cte_segment.get_identifier().raw == "cte"

    # BaseFileSegment.get_table_references & StatementSegment.get_table_references
    assert parsed.tree.get_table_references() == {"tab_a", "tab_b"}


def test_normalise_newlines():
    """Test normalising newlines to unix-style line endings."""
    in_str = "SELECT\r\n foo\n FROM \r \n\r bar;"
    out_str = "SELECT\n foo\n FROM \n \n\n bar;"
    assert out_str == Linter._normalise_newlines(in_str)


@pytest.mark.parametrize(
    "fix_even_unparsable",
    [False, True],
)
def test_unparsable_fix_output(fix_even_unparsable):
    """Tests functionality and logging output with unparsable sections.

    NOTE: While we cover different paths, the result for this test is the
    same for both values of `fix_even_unparsable`. We probably need a better
    test case at some point so that we can actually see the difference.
    """
    config = FluffConfig(
        overrides={"fix_even_unparsable": fix_even_unparsable, "dialect": "ansi"}
    )
    linter = Linter(config=config)
    # Attempt to fix it, capturing the logging output.
    with fluff_log_catcher(logging.WARNING, "sqlfluff.linter") as caplog:
        result = linter.lint_paths(
            ("test/fixtures/linter/parse_error_2.sql",),
            fix=True,
            apply_fixes=True,
            fixed_file_suffix=f"_{fix_even_unparsable}_fix",
            fix_even_unparsable=fix_even_unparsable,
        )
    # Assert that it parsed (i.e. we found a select_statement), but with an
    # unparsable section in there too.
    assert result.tree
    assert "select_statement" in result.tree.descendant_type_set
    assert "unparsable" in result.tree.descendant_type_set
    # We should still find linting issues too
    assert result.check_tuples(raise_on_non_linting_violations=False) == [
        ("CP01", 2, 7),  # `a as b` - capitalisation of AS
        ("AL03", 3, 5),  # 42 is an expression without an alias
        # The unparsable section is (wrongly) detected as an indentation issue.
        ("LT02", 4, 1),
        ("CP01", 5, 1),  # `from` is uncapitalised
    ]
    # We should make sure that the warning that asks users to report a bug is
    # NOT present. i.e. the warning which could happen in `lint_fix_parsed()`.`
    assert "Please report this as a bug" not in caplog.text
    # Also not the `fix not applied`. The one in `_warn_unfixable()`
    assert "it would re-cause the same error" not in caplog.text
    # In fact, there shouldn't be any warnings at all.
    assert not caplog.text.strip()
    # In both cases, the final capitalisation and the `a as b` sections should have
    # been fixed (because they aren't in the unparsable section).
    assert "from cte" not in result.tree.raw
    assert "FROM cte" in result.tree.raw
    assert "a as b" not in result.tree.raw
    assert "a AS b" in result.tree.raw
    # Check whether the file was persisted. If `fix_even_unparsable` was set, then
    # there should be a file, and it should have the fixes from above in it. If not
    # then there should be no fixed file, as the persist will have been aborted due
    # to the parsing issues.
    predicted_fix_path = (
        f"test/fixtures/linter/parse_error_2_{fix_even_unparsable}_fix.sql"
    )
    if fix_even_unparsable:
        with open(predicted_fix_path, "r") as f:
            fixed_sql = f.read()
        assert result.tree.raw == fixed_sql
    else:
        with pytest.raises(FileNotFoundError):
            open(predicted_fix_path, "r")



"""Lower level routines for config file loading and caching.

Functions in this module load config from *individual* files and
resources. While some are cached, they are cached on the basis of
not processing individual files more than once.

For the cached functions it is VERY recommended to make sure they
are copied before any edits happen to them, as those edits may
propagate back up into the cache. Typically the results are passed
to `nested_combine` either immediately, or eventually after returning
which should negate this effect.
"""

import os.path
import sys
from typing import Optional

from sqlfluff.core.config.ini import load_ini_string
from sqlfluff.core.config.toml import load_toml_file_config
from sqlfluff.core.config.validate import validate_config_dict
from sqlfluff.core.helpers.string import (
    split_comma_separated_string,
)
from sqlfluff.core.types import ConfigMappingType

if sys.version_info >= (3, 9):
    # Explicitly rename so that mypy is happy re-exporting it for other
    # modules to use.
    from functools import cache as cache
else:  # pragma: no cover
    from functools import lru_cache

    # With maxsize set to `None`, the lru_cache approximates what the later
    # introduced `cache` does. We don't need to worry too much about overflow
    # as config files are usually small, and sqlfluff is not often a long
    # lived process.
    cache = lru_cache(maxsize=None)


COMMA_SEPARATED_PATH_KEYS = (
    "load_macros_from_path",
    "loader_search_path",
    "exclude_macros_from_path",
)
RESOLVE_PATH_SUFFIXES = ("_path", "_dir")


def _load_raw_file_as_dict(filepath: str) -> ConfigMappingType:
    """Loads the raw dict object from file without interpolation."""
    filename = os.path.basename(filepath)
    if filename == "pyproject.toml":
        return load_toml_file_config(filepath)
    # If it's not a pyproject file, assume that it's an ini file.
    with open(filepath, mode="r") as file:
        return load_ini_string(file.read())


def _resolve_path(filepath: str, val: str) -> str:
    """Try to resolve a path found in a config value."""
    # Make the referenced path.
    ref_path = os.path.join(os.path.dirname(filepath), val)
    # Check if it exists, and if it does, replace the value with the path.
    return ref_path if os.path.exists(ref_path) else val


def _resolve_paths_in_config(
    config: ConfigMappingType, filepath: str, logging_reference: Optional[str] = None
) -> None:
    """Attempt to resolve any paths found in the config file.

    NOTE: This method is recursive to crawl the whole config object,
    and also mutates the underlying config object rather than returning it.
    """
    log_filename: str = logging_reference or filepath
    for key, val in config.items():
        # If it's a dict, recurse.
        if isinstance(val, dict):
            _resolve_paths_in_config(val, filepath, logging_reference=logging_reference)
        # If it's a potential multi-path, split, resolve and join
        if key.lower() in COMMA_SEPARATED_PATH_KEYS:
            assert isinstance(
                val, str
            ), f"Value for {key} in {log_filename} must be a string not {type(val)}."
            paths = split_comma_separated_string(val)
            config[key] = ",".join(_resolve_path(filepath, path) for path in paths)
        # It it's a single path key, resolve it.
        elif key.lower().endswith(RESOLVE_PATH_SUFFIXES):
            assert isinstance(
                val, str
            ), f"Value for {key} in {log_filename} must be a string not {type(val)}."
            config[key] = _resolve_path(filepath, val)


@cache
def load_config_file_as_dict(filepath: str) -> ConfigMappingType:
    """Load the given config file into a dict and validate.

    This method is cached to mitigate being called multiple times.

    This doesn't manage the combination of config files within a nested
    structure, that happens further up the stack.
    """
    raw_config = _load_raw_file_as_dict(filepath)

    # The raw loaded files have some path interpolation which is necessary.
    _resolve_paths_in_config(raw_config, filepath)
    # Validate
    validate_config_dict(raw_config, filepath)

    # Return dict object (which will be cached)
    return raw_config


@cache
def load_config_string_as_dict(
    config_string: str, working_path: str, logging_reference: str
) -> ConfigMappingType:
    """Load the given config string and validate.

    This method is cached to mitigate being called multiple times.

    This doesn't manage the combination of config files within a nested
    structure, that happens further up the stack. The working path is
    necessary to resolve any paths in the config file.
    """
    raw_config = load_ini_string(config_string)

    # The raw loaded files have some path interpolation which is necessary.
    _resolve_paths_in_config(
        raw_config, working_path, logging_reference=logging_reference
    )
    # Validate
    validate_config_dict(raw_config, logging_reference)

    # Return dict object (which will be cached)
    return raw_config



"""Tests the python routines within LT02 and LT04."""

import pytest

from sqlfluff.core import FluffConfig, Linter


@pytest.mark.parametrize(
    ["in_sql", "out_sql"],
    [
        (
            """SELECT
    acct_id,
    date_x,
    't' AS test,

    CASE
        WHEN condition_1 = '1' THEN ''
        ELSE condition_1
    END AS case_1,

    CASE
        WHEN condition_2 = '2' THEN ''
        ELSE condition_2
    END AS case_2,
    dollar_amt,
FROM
    table_x""",
            """SELECT
    acct_id
    , date_x
    , 't' AS test

    , CASE
        WHEN condition_1 = '1' THEN ''
        ELSE condition_1
    END AS case_1

    , CASE
        WHEN condition_2 = '2' THEN ''
        ELSE condition_2
    END AS case_2
    , dollar_amt,
FROM
    table_x""",
        ),
    ],
)
def test_rules_std_LT02_LT04_interaction_indentation_leading(in_sql, out_sql) -> None:
    """Test interaction between LT02 and LT04.

    Test sql with two newlines with trailing commas expecting leading.
    """
    # Lint expected rules.
    cfg = FluffConfig.from_string(
        """[sqlfluff]
dialect = snowflake
rules = LT02, LT04

[sqlfluff:layout:type:comma]
spacing_before = touch
line_position = leading
"""
    )
    linter = Linter(config=cfg)

    # Return linted/fixed file.
    linted_file = linter.lint_string(in_sql, fix=True)

    # Check expected lint errors are raised.
    assert set([v.rule.code for v in linted_file.violations]) == {"LT04"}

    # Check file is fixed.
    assert linted_file.fix_string()[0] == out_sql



"""An example to show a few ways of configuring the API."""

import sqlfluff
from sqlfluff.core import FluffConfig, Linter

# #######################################
# The simple API can be configured in three ways.

# 1. Limited keyword arguments
sqlfluff.fix("SELECT  1", dialect="bigquery")

# 2. Providing the path to a config file
sqlfluff.fix("SELECT  1", config_path="test/fixtures/.sqlfluff")

# 3. Providing a preconfigured FluffConfig object.
# NOTE: This is the way of configuring SQLFluff which will give the most control.

# 3a. FluffConfig objects can be created directly from a dictionary of values.
config = FluffConfig(configs={"core": {"dialect": "bigquery"}})
# 3b. FluffConfig objects can be created from a config file in a string.
config = FluffConfig.from_string("[sqlfluff]\ndialect=bigquery\n")
# 3c. FluffConfig objects can be created from a config file in multiple strings
#     to simulate the effect of multiple nested config strings.
config = FluffConfig.from_strings(
    # NOTE: Given these two strings, the resulting dialect would be "mysql"
    # as the later files take precedence.
    "[sqlfluff]\ndialect=bigquery\n",
    "[sqlfluff]\ndialect=mysql\n",
)
# 3d. FluffConfig objects can be created from a path containing a config file.
config = FluffConfig.from_path("test/fixtures/")
# 3e. FluffConfig objects can be from keyword arguments
config = FluffConfig.from_kwargs(dialect="bigquery", rules=["LT01"])

# The FluffConfig is then provided via a config argument.
sqlfluff.fix("SELECT  1", config=config)


# #######################################
# The core API is always configured using a FluffConfig object.

# When instantiating a Linter (or Parser), a FluffConfig must be provided
# on instantiation. See above for details on how to create a FluffConfig.
linter = Linter(config=config)

# The provided config will then be used in any operations.

lint_result = linter.lint_string("SELECT  1", fix=True)
fixed_string = lint_result.fix_string()
# NOTE: The "True" element shows that fixing was a success.
assert fixed_string == ("SELECT 1", True)



"""Tests issue #1373 doesn't reoccur.

The combination of LT02 (incorrect indentation), LT09 (select targets),
and LT01 (unnecessary white space) can result in incorrect indentation.
"""

import sqlfluff


def test__rules__std_LT02_LT09_LT01():
    """Verify that double indents don't flag LT01."""
    sql = """
WITH example AS (
    SELECT my_id,
        other_thing,
        one_more
    FROM
        my_table
)

SELECT my_id
FROM example\n"""
    fixed_sql = """
WITH example AS (
    SELECT
        my_id,
        other_thing,
        one_more
    FROM
        my_table
)

SELECT my_id
FROM example\n"""
    result = sqlfluff.fix(sql, exclude_rules=["LT13"])
    assert result == fixed_sql



"""Tests for the BaseGrammar and it's methods.

NOTE: All of these tests depend somewhat on the KeywordSegment working as planned.
"""

import pytest

from sqlfluff.core.dialects.base import Dialect
from sqlfluff.core.errors import SQLParseError
from sqlfluff.core.parser import (
    CodeSegment,
    KeywordSegment,
    StringParser,
    SymbolSegment,
    WhitespaceSegment,
)
from sqlfluff.core.parser.context import ParseContext
from sqlfluff.core.parser.lexer import RegexLexer
from sqlfluff.core.parser.match_algorithms import (
    greedy_match,
    next_ex_bracket_match,
    next_match,
    resolve_bracket,
    trim_to_terminator,
)

# NB: All of these tests depend somewhat on the KeywordSegment working as planned


@pytest.fixture(scope="function")
def test_dialect():
    """A stripped back test dialect for testing brackets."""
    test_dialect = Dialect("test", root_segment_name="FileSegment")
    test_dialect.bracket_sets("bracket_pairs").update(
        [("round", "StartBracketSegment", "EndBracketSegment", True)]
    )
    test_dialect.set_lexer_matchers(
        [
            RegexLexer("whitespace", r"[^\S\r\n]+", WhitespaceSegment),
            RegexLexer(
                "code", r"[0-9a-zA-Z_]+", CodeSegment, segment_kwargs={"type": "code"}
            ),
        ]
    )
    test_dialect.add(
        StartBracketSegment=StringParser("(", SymbolSegment, type="start_bracket"),
        EndBracketSegment=StringParser(")", SymbolSegment, type="end_bracket"),
    )
    # Return the expanded copy.
    return test_dialect.expand()


def make_result_tuple(result_slice, matcher_keywords, test_segments):
    """Make a comparison tuple for test matching."""
    # No result slice means no match.
    if not result_slice:
        return ()

    return tuple(
        (
            KeywordSegment(elem.raw, pos_marker=elem.pos_marker)
            if elem.raw in matcher_keywords
            else elem
        )
        for elem in test_segments[result_slice]
    )


@pytest.mark.parametrize(
    "matcher_keywords,result_slice,winning_matcher",
    [
        # Basic version, we should find bar first
        (["bar", "foo"], slice(0, 1), "bar"),
        # Look ahead for foo
        (["foo"], slice(2, 3), "foo"),
        # Duplicate matchers
        (["foo", "foo"], slice(2, 3), "foo"),
        (["sadkjfhas", "asefaslf"], slice(0, 0), None),
    ],
)
def test__parser__algorithms__next_match(
    matcher_keywords,
    result_slice,
    winning_matcher,
    test_segments,
):
    """Test the `next_match()` method."""
    # Make the string parsers for testing.
    matchers = [StringParser(keyword, KeywordSegment) for keyword in matcher_keywords]
    # Fetch the matching keyword from above (because it will have the same position)
    if winning_matcher:
        winning_matcher = matchers[matcher_keywords.index(winning_matcher)]

    ctx = ParseContext(dialect=None)
    match, matcher = next_match(
        test_segments,
        0,
        matchers,
        ctx,
    )

    # Check the right matcher was successful.
    if winning_matcher:
        assert matcher is winning_matcher
    else:
        # If no designated winning matcher, assert that it wasn't successful.
        assert matcher is None
        assert not match
    assert match.matched_slice == result_slice


@pytest.mark.parametrize(
    "raw_segments,result_slice,error",
    [
        (["(", "a", ")", " ", "foo"], slice(0, 3), None),
        (["(", "a", "(", "b", ")", "(", "c", ")", "d", ")", "e"], slice(0, 10), None),
        # This should error because we try to close a square bracket
        # inside a round one.
        (["(", "a", "]", "b", ")", "e"], None, SQLParseError),
        # This should error because we never find the end.
        (["(", "a", " ", "b", " ", "e"], None, SQLParseError),
    ],
)
def test__parser__algorithms__resolve_bracket(
    raw_segments, result_slice, error, generate_test_segments
):
    """Test the `resolve_bracket()` method."""
    test_segments = generate_test_segments(raw_segments)
    start_bracket = StringParser("(", SymbolSegment, type="start_bracket")
    end_bracket = StringParser(")", SymbolSegment, type="end_bracket")
    start_sq_bracket = StringParser("[", SymbolSegment, type="start_square_bracket")
    end_sq_bracket = StringParser("]", SymbolSegment, type="end_square_bracket")
    ctx = ParseContext(dialect=None)

    # For this test case we assert that the first segment is the initial match.
    first_match = start_bracket.match(test_segments, 0, ctx)
    assert first_match

    args = (test_segments,)
    kwargs = dict(
        opening_match=first_match,
        opening_matcher=start_bracket,
        start_brackets=[start_bracket, start_sq_bracket],
        end_brackets=[end_bracket, end_sq_bracket],
        bracket_persists=[True, False],
        parse_context=ctx,
    )
    # If an error is defined, check that it is raised.
    if error:
        with pytest.raises(error):
            resolve_bracket(*args, **kwargs)
    else:
        result = resolve_bracket(*args, **kwargs)
        assert result
        assert result.matched_slice == result_slice


@pytest.mark.parametrize(
    "raw_segments,target_word,result_slice",
    [
        ([], "foo", slice(0, 0)),
        (["(", "foo", ")", " ", "foo"], "foo", slice(4, 5)),
        (["a", " ", "foo", " ", "foo"], "foo", slice(2, 3)),
        (["foo", " ", "foo", " ", "foo"], "foo", slice(0, 1)),
        # Error case, unexpected closing bracket.
        # NOTE: This should never normally happen, but we should
        # be prepared in case it does so that we return appropriately.
        (["a", " ", ")", " ", "foo"], "foo", slice(0, 0)),
    ],
)
def test__parser__algorithms__next_ex_bracket_match(
    raw_segments, target_word, result_slice, generate_test_segments, test_dialect
):
    """Test the `next_ex_bracket_match()` method."""
    test_segments = generate_test_segments(raw_segments)
    target = StringParser(target_word, KeywordSegment)
    ctx = ParseContext(dialect=test_dialect)

    result, _, _ = next_ex_bracket_match(
        test_segments,
        0,
        matchers=[target],
        parse_context=ctx,
    )

    assert result.matched_slice == result_slice


@pytest.mark.parametrize(
    "raw_segments,target_words,inc_term,result_slice",
    [
        (["a", "b", " ", "c", "d", " ", "e"], ["e", "c"], False, slice(0, 2)),
        (["a", "b", " ", "c", "d", " ", "e"], ["e", "c"], True, slice(0, 4)),
        # NOTE: Because "b" is_alpha, it needs whitespace before it to match.
        (["a", "b", " ", "b"], ["b"], True, slice(0, 4)),
        (["a", "b", " ", "b"], ["b"], False, slice(0, 2)),
        (["a", "b", "c", " ", "b"], ["b"], False, slice(0, 3)),
    ],
)
def test__parser__algorithms__greedy_match(
    raw_segments,
    target_words,
    inc_term,
    result_slice,
    generate_test_segments,
    test_dialect,
):
    """Test the `greedy_match()` method."""
    test_segments = generate_test_segments(raw_segments)
    matchers = [StringParser(word, KeywordSegment) for word in target_words]
    ctx = ParseContext(dialect=test_dialect)

    match = greedy_match(
        segments=test_segments,
        idx=0,
        parse_context=ctx,
        matchers=matchers,
        include_terminator=inc_term,
    )

    assert match
    assert match.matched_slice == result_slice


@pytest.mark.parametrize(
    "raw_segments,target_words,expected_result",
    [
        # Terminators mid sequence.
        (["a", "b", " ", "c", "d", " ", "e"], ["e", "c"], 2),
        # Initial terminators.
        (["a", "b", " ", "c", "d", " ", "e"], ["a", "e"], 0),
        # No terminators.
        (["a", "b", " ", "c", "d", " ", "e"], ["x", "y"], 7),
        # No sequence.
        ([], ["x", "y"], 0),
    ],
)
def test__parser__algorithms__trim_to_terminator(
    raw_segments,
    target_words,
    expected_result,
    generate_test_segments,
    test_dialect,
):
    """Test the `trim_to_terminator()` method."""
    test_segments = generate_test_segments(raw_segments)
    matchers = [StringParser(word, KeywordSegment) for word in target_words]
    ctx = ParseContext(dialect=test_dialect)

    assert (
        trim_to_terminator(
            segments=test_segments,
            idx=0,
            parse_context=ctx,
            terminators=matchers,
        )
        == expected_result
    )



"""Tests observed conflict between ST05 & LT09."""

from sqlfluff.core import FluffConfig, Linter


def test__rules__std_ST05_LT09_4137() -> None:
    """Tests observed conflict between ST05 & LT09.

    In this case, the moved `t2` table was created after the first usage.
    https://github.com/sqlfluff/sqlfluff/issues/4137
    """
    sql = """
with

cte1 as (
    select t1.x, t2.y
    from tbl1 t1
    join (select x, y from tbl2) t2
        on t1.x = t2.x
)

, cte2 as (
    select x, y from tbl2 t2
)

select x, y from cte1
union all
select x, y from cte2
;
"""
    fixed_sql = """
with t2 as (select
x,
y
from tbl2),
cte1 as (
    select
t1.x,
t2.y
    from tbl1 t1
    join t2
        on t1.x = t2.x
),
cte2 as (
    select
x,
y
from tbl2 t2
)
select
x,
y
from cte1
union all
select
x,
y
from cte2
;
"""
    cfg = FluffConfig.from_kwargs(
        dialect="ansi",
        rules=["ST05", "LT09"],
    )
    result = Linter(config=cfg).lint_string(sql, fix=True)
    assert result.fix_string()[0] == fixed_sql


def test__rules__std_ST05_LT09_5265() -> None:
    """Tests observed conflict between ST05 & LT09.

    In this case, the moved `t2` table was created after the first usage.
    https://github.com/sqlfluff/sqlfluff/issues/4137
    """
    sql = """
with

cte1 as (
    select t1.x, t2.y
    from tbl1 t1
    join (select x, y from tbl2) t2
        on t1.x = t2.x
)

, cte2 as (
    select x, y from tbl2 t2
)

select x, y from cte1
union all
select x, y from cte2
;
"""
    fixed_sql = """
with t2 as (select
x,
y
from tbl2),
cte1 as (
    select
t1.x,
t2.y
    from tbl1 t1
    join t2
        on t1.x = t2.x
),
cte2 as (
    select
x,
y
from tbl2 t2
)
select
x,
y
from cte1
union all
select
x,
y
from cte2
;
"""
    cfg = FluffConfig.from_kwargs(
        dialect="ansi",
        rules=["ST05", "LT09"],
    )
    result = Linter(config=cfg).lint_string(sql, fix=True)
    assert result.fix_string()[0] == fixed_sql
